/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 *    AprioriSetsAndSequences.java 
 *    by Keith A. Pray, Worcester Polytechnic Institute, 2002
 *
 *      Based on 
 *      AprioriSets.java 
 *        by Zack Stoecker-Sylvia, Worcester Polytechnic Institute, 2002
 *      Apriori.java, Copyright (C) 1999 Eibe Frank,Mark Hall
 *      and 
 *      ARMiner, Copyright (C) 2000  UMass/Boston - Computer Science Department
 */

package wpi.associations;

import java.io.*;
import java.util.*;

import weka.core.*;
import weka.filters.Filter;
import weka.filters.AttributeFilter;

import wpi.timesequence.*;
import wpi.associations.arminerSequence.*;

/*
  Maintainance log started on April 30th, 2002

  April 30th, 2002   - added support for time sequence attributes,
                     - cleaned up formatting -kap

  May 17th, 2002     - implemented main method -kap

  June 11th, 2002    - added support for loading frequent itemsets
                       from a specified cache file -kap

  January 13th, 2003 - changed buildAssociations to mineAssociations
                       so that we could override the buildAssociations
		       method in our own abstract Associator class.
		       This facilitated the measuring of mining time
		       across all our association rule mining classes. -kap

  February 10, 2003  - removed verbose option, no longer functional
                       for this implementation. see AprioriSets for
		       removed code. -kap

		     - removed data member FastVector m_Ls, the set of
		       all sets of itemsets L, FastVector m_hashtables,
		       the same information stored in hash tables, 
		       FastVector[] m_allTheRules, the list of all generated
		       rules\, and boolean m_outputItemSets, an option to
		       output itemsets found.
		       these are covered by ARMinerApriori now. 
		       I don't know why Zach left them.
		       see AprioriSets for removed code. -kap
		       
		     - added -W maxEvents to option reporting methods -kap

  March 12, 2003     - added support and confidence message when building with
                       cache file

		     - added "Beginning to mine..." message. was in 
		       ARMinerApriori.java but moved it for consistency. -kap

  August 13, 2003    - added Lift and Chi-Square to rule reporting -kap

  August 14, 2003    - added log file option, logStats data member,
                       set and get methods, tip text method, etc. -kap

  January 12, 2004   - added logging of NUM_EVENTS_OF_SAME_TYPE_ALLOWED
                       to mineAssociations method -kap

  March 15, 2004     - added event confidence status reporting -kap
*/

/**
 * Class implementing an Apriori-type algorithm that may understand sets.  
 * Sets must be loaded as strings into WEKA and are then parsed using ^ or &
 * to delimit items.  Items are split into individual Attribute-Value pairs 
 * for the item based ARMiner algorithm.  WEKA iteratively reduces the minimum
 * support until it finds the required number of rules with the given minimum 
 * confidence. <p>
 *
 * This class implements the Apriori algorithm 
 * for finding large itemsets.
 *
 * (see "Fast Algorithms for Mining Association Rules"
 * by Rakesh Agrawal and Ramakrishnan Srikant
 * from IBM Almaden Research Center 1994)
 *
 * Valid options are:<p>
 *
 * -N required number of rules <br>
 * The required number of rules (default: 10). <p>
 * 
 * -C minimum confidence score of a rule <br>
 * The minimum confidence of a rule (default: 0.9). <p>
 * 
 * -D delta for minimum support <br>
 * The delta by which the minimum support is decreased in
 * each iteration (default: 0.05).
 * 
 * -U upper bound for minimum support <br>
 * The upper bound for minimum support. Don't explicitly look for 
 * rules with more than this level of support. <p>
 * 
 * -M lower bound for minimum support <br>
 * The lower bound for the minimum support (default = 0.1). <p>
 * 
 * -A <br>
 * The range of required attributes for the antecedents. <p>
 * 
 * -B <br>
 * The maximum number of antecedent items. <p>
 * 
 * -W <br>
 * Maximum number of event items of the same type allowed in a rule. <p>
 *
 * -Y <br>
 * The range of required attributes for the consequents. <p>
 * 
 * -Z <br>
 * The minimum number of consequent items. <p>
 * 
 * -F <br>
 * The cache of frequent itemsets to use.
 *
 * -L <br>
 * Write statistics to a log file. <p>
 *
 * @author Eibe Frank (eibe@cs.waikato.ac.nz)
 * @author Mark Hall (mhall@cs.waikato.ac.nz)
 * @author Zack Stoecker-Sylvia (zss@wpi.edu)
 * @author Keith A. Pray (kap@wpi.edu)
 * @version 1.11
 */

public class AprioriSetsAndSequences extends Associator 
  implements OptionHandler 
{

//********************************************************//
// ----- Data Members -----
//********************************************************//  

  /** The minimum support. */
  protected double m_minSupport;
  
  /** The upper bound on the support */
  protected double m_upperBoundMinSupport;
  
  /** The lower bound for the minimum support. */
  protected double m_lowerBoundMinSupport;
  
  /** The minimum confidence score. */
  protected double m_minConfidence;
  
  /** The maximum number of rules that are output. */
  protected int m_maxRulesOutput;
  
  /** Delta by which m_minSupport is decreased in each iteration. */
  protected double m_delta;
  
  /** Number of cycles used before required number of rules was one. */
  protected int m_cycles;
  
  /** The instances (transactions) to be used for generating 
   *  the association rules. */
  protected Instances m_instances;
  
  /* -- zss: New options -- */

  // kap added
  /** The minimum number of antecedents in a rule */
  protected int m_minAntecedent;

  /** The maximum number of antecedents in a rule */
  protected int m_maxAntecedent;
  
  /** The minimum number of consequents in a rule */
  protected int m_minConsequent;
  
  // kap added
  /** The maximum number of consequents in a rule */
  protected int m_maxConsequent;

  // kap added
  /** The maximum number of events of the same kind allowed in a rule. */
  protected int maxEvents;
  
  /** The vector containing all found AssociationRules */
  protected Vector m_rulesFound;
  
  /** The attributes required to be in the antecedents of the rules */
  protected Range m_requiredAntecedentAttributes = new Range();
  
  /** The attributes required to be in the consequents of the rules */
  protected Range m_requiredConsequentAttributes = new Range();
  
  /** A hash listing attribute value pair Strings by Integer key */
  protected Hashtable m_attributeHash;

  /** File name to load cached frequent itemsets from */
  protected String cacheFileName;

  /** Specifies statistics should be written to a log file. */
  protected boolean logStats;

  /** Specifies debug info level 
   * 0: no debug info
   * 1: input to methods
   * 2: and output from methods
   * 3: and all sorts of stuff 
   */
  private static final int debug = 0;

//********************************************************//
// ----- Methods-----
//********************************************************//
  
  /**
   * Returns a string describing this associator
   * @return a description of the evaluator suitable for
   * displaying in the explorer/experimenter gui
   */

  public String globalInfo()
  {
    return ( "Finds association rules using an extended Apriori " +
	     "algorithm that can handle set and time sequence attributes." );

  } // END public String globalInfo()

//********************************************************//
  
  /**
   * Constructor that allows to sets default values for the 
   * minimum confidence and the maximum number of rules
   * the minimum confidence.
   */

  public AprioriSetsAndSequences() 
  {
    resetOptions();

  } // END public AprioriSetsAndSequences() 

//********************************************************//
  
  /**
   * Resets the options to the default values.
   */

  public void resetOptions() 
  {
    m_delta = 0.05;
    m_minConfidence = 0.90;
    m_maxRulesOutput = 10;
    m_lowerBoundMinSupport = 0.1;

    // this should be 0.95. it seems the only reason it was 1
    // was because the code initialized the support to this - m_delta
    // which cause mining to start with a support less than that 
    // specified by the user. -kap
    //    m_upperBoundMinSupport = 1.0;
    m_upperBoundMinSupport = 0.95;

    // otherwise we would get many many I think
    maxEvents = 2;
    
    m_requiredAntecedentAttributes.setRanges ( "" );
    m_requiredConsequentAttributes.setRanges ( "" );
    m_minAntecedent = 0;
    m_maxAntecedent = 0;
    m_minConsequent = 0;
    m_maxConsequent = 0;

    // we would normally not load frequent itemsets from
    // cache BEFORE mining for them -kap
    cacheFileName = "";

    // log file turned off by default
    logStats = false;
    logStatsFileName = "";

  } // END public void resetOptions() 

//********************************************************//

  /* 
     If a cache file name is specified then the instances
     specified will not be used to generate rules. The
     itemsets stored in the cache will be used instead.
     Please note that the cache file should be the result
     of mining frequent itemsets from the instances specified.

     Changing the method below to be named mineAssociations
     instead of buildAssociations. This was done so that we
     could measure the time it takes to mine rules in the
     abstract class that this class extends.
     -kap
  */
  
  /**
   * Method that generates all large itemsets with a
   * minimum support, and from these all association
   * rules with a minimum confidence.
   * 
   * @param instances
   * the instances to be used for generating the associations
   * 
   * @exception	Exception
   * if rules can't be built successfully
   */

  public void mineAssociations ( Instances instances ) throws Exception 
  {
    m_instances = instances;
    m_requiredAntecedentAttributes.setUpper ( instances.numAttributes() - 1 );
    m_requiredConsequentAttributes.setUpper ( instances.numAttributes() - 1 );

    DBReader instanceReader = new DBReader ( instances );

    m_attributeHash = instanceReader.getNumberHash();

    int necSupport = 0;

    Vector requiredAntecedents = new Vector();
    Vector requiredConsequents = new Vector();

    ARMinerApriori itemsetMiner = new ARMinerApriori();
    AprioriRules rulesFinder = new AprioriRules();

    // tell the itemset miner the maximum number of events of the same
    // kind to allow in a rule
    itemsetMiner.setMaxEvents ( maxEvents );

    // if we have a logger tell the item set miner about it
    if ( logger != null )
    {
      itemsetMiner.setLogger ( logger );

      // while we're at it let's log some stuff since we
      // know there's a logger
      logger.logValue ( Logger.NUM_EVENTS_OF_SAME_TYPE_ALLOWED,
			Integer.toString ( maxEvents ) );

      logger.logValue ( Logger.CONFIDENCE,
			Double.toString ( m_minConfidence ) );

    } // end if loggin is on
    
    // we must pass the m_attributeHash ( aka number hash )
    // to the new AprioriRules ( aka rulesFinder ) to support time sequence
    // attributes -kap
    rulesFinder.setNumberHash ( m_attributeHash );

    // let's also tell the rules finder how many instances
    // there are so nifty statistical measures like chi-square
    // can be computed -kap 2003-08-13
    rulesFinder.setNumInstances ( instances.numInstances() );

    // build the list of required items 
    for ( int i = 0; i < instances.numAttributes(); i++ )
    {
      if ( m_requiredAntecedentAttributes.isInRange ( i ) )
      {
	requiredAntecedents.
	  addElement ( instanceReader.getAttributeItemset ( i ) );
	
	switch ( debug ) {
	case 3: // case 0:
	  System.out.println ( "AprioriSetsAndSequences.mineAssociations: " +
			       "added itemset= " + 
			       requiredAntecedents.lastElement() + " " +
			       "to requiredAntecedents." );
	} // end switch debug
	
      } // end if attribute in range of required antecedents
      
      if ( m_requiredConsequentAttributes.isInRange ( i ) )
      {
	requiredConsequents.
	  addElement ( instanceReader.getAttributeItemset ( i ) );
	
	switch ( debug ) {
	case 3: // case 0:
	  System.out.println ( "AprioriSetsAndSequences.mineAssociations: " +
			       "added itemset= " + 
			       requiredConsequents.lastElement() + " " +
			       "to requiredConsequents." );
	} // end switch debug
	
      } // end if attribute in range of required consequents
      
    } // end for all the attributes, build required items list
    
    if ( requiredAntecedents.isEmpty() )
    {
      requiredAntecedents = null;
    }
    
    if ( requiredConsequents.isEmpty() )
    {
      requiredConsequents = null;
    }
    
    // Decrease minimum support until desired number of rules found.
    m_cycles = 0;

    // I don't think this should be here... it has the result of starting 
    // with support m_delta less than what the user entered -kap

    //    m_minSupport = m_upperBoundMinSupport - m_delta;

    m_minSupport = m_upperBoundMinSupport;

    m_minSupport = ( m_minSupport < m_lowerBoundMinSupport )
      ? m_lowerBoundMinSupport 
      : m_minSupport;

    // let's see how long it takes to generate the rules
    long startTime, endTime, secondsGeneration, secondsConfidence;

    do 
    {
      // if we loaded itemset from a cache file don't find large itemsets
      if ( cacheFileName.length() == 0 )
      {
	// report status
	System.out.println ( "Beginning to mine with support " + 
			     m_minSupport +
			     ", confidence " + m_minConfidence + "...");

	// find the frequent itemsets -kap
	itemsetMiner.findFrequentItemsets 
	  ( instanceReader, new DBCacheWriter ( "temp.cache" ),
	    (float)m_minSupport );

	// let's see how long it takes to generate the rules
	startTime = System.currentTimeMillis();

	// construct the rules -kap
	m_rulesFound = rulesFinder.findAssociations 
	  ( new DBCacheReader ( "temp.cache" ),
	    (float)m_minSupport,
	    (float)m_minConfidence,
	    requiredAntecedents,
	    requiredConsequents,
	    m_minAntecedent,
	    m_maxAntecedent,
	    m_minConsequent,
	    m_maxConsequent );

	endTime = System.currentTimeMillis();
    
	secondsGeneration = ( endTime / 1000 ) - ( startTime / 1000 );

	System.out.println ( "\nAprioriSetsAndSequences.mineAssociations: " + 
			     "rule generation took " + 
			     secondsGeneration + " seconds." );

	// let's see how long it takes to find confidence for each
	// rule
	startTime = System.currentTimeMillis();

	// have to check the confidence of rules with events
	// seperately. This is done here because the data set 
	// is needed to calculate the confidence. -kap 2003-11-26
	m_rulesFound = eventConfidence ( m_rulesFound, instanceReader, 
					 (float)m_minConfidence );

	endTime = System.currentTimeMillis();

	secondsConfidence = ( endTime / 1000 ) - ( startTime / 1000 );

	System.out.println ( "\nrule confidence finding took " + 
			     secondsConfidence + " seconds. " +
			     m_rulesFound.size() + " rules left." );

	// if we are logging, log these
	if ( logger != null )
	{
	  logger.logValue ( Logger.SECONDS_TO_GENERATE_RULES,
			    Long.toString ( secondsGeneration ) );
	  logger.logValue ( Logger.SECONDS_TO_FIND_CONFIDENCE,
			    Long.toString ( secondsConfidence ) );

	  // might as well log number of rules found too
	  logger.logValue ( Logger.NUM_RULES,
			    Integer.toString ( m_rulesFound.size() ) );

	  // log the support used while we are at it
	  logger.logValue ( Logger.SUPPORT,
			    Double.toString ( m_minSupport ) );

	} // end if we are logging

      } // end if cache file name not specified
      else
      {
	// all we have to do is tell the rules finder to use the specfied
	// cache file
	
	// might as well print out a status message for those terminal types
	System.out.println ( "Building rules with cache file: " +
			     cacheFileName + ", and support " +
			     + m_minSupport + 
			     ", confidence = " + m_minConfidence );

	// let's see how long it takes to generate the rules
	startTime = System.currentTimeMillis();

	m_rulesFound = rulesFinder.findAssociations 
	  ( new DBCacheReader ( cacheFileName ),
	    (float)m_minSupport,
	    (float)m_minConfidence,
	    requiredAntecedents,
	    requiredConsequents,
	    m_minAntecedent,
	    m_maxAntecedent,
	    m_minConsequent,
	    m_maxConsequent );

	endTime = System.currentTimeMillis();
    
	secondsGeneration = ( endTime / 1000 ) - ( startTime / 1000 );

	System.out.println ( "\nAprioriSetsAndSequences.mineAssociations: " + 
			     "rule generation took " + 
			     secondsGeneration + " seconds." );

	// let's see how long it takes to find confidence for each
	// rule
	startTime = System.currentTimeMillis();
	
	// have to check the confidence of rules with events
	// seperately. This is done here because the data set 
	// is needed to calculate the confidence. -kap 2003-11-26
	m_rulesFound = eventConfidence ( m_rulesFound, instanceReader,
					 (float)m_minConfidence );

	endTime = System.currentTimeMillis();

	secondsConfidence = ( endTime / 1000 ) - ( startTime / 1000 );

	System.out.println ( "\nrule confidence finding took " + 
			     secondsConfidence + " seconds. " +
			     m_rulesFound.size() + " rules left." );
	// if we are logging, log these
	if ( logger != null )
	{
	  logger.logValue ( Logger.SECONDS_TO_GENERATE_RULES,
			    Long.toString ( secondsGeneration ) );
	  logger.logValue ( Logger.SECONDS_TO_FIND_CONFIDENCE,
			    Long.toString ( secondsConfidence ) );

	  // might as well log number of rules found too
	  logger.logValue ( Logger.NUM_RULES,
			    Integer.toString ( m_rulesFound.size() ) );

	  // log the support used while we are at it
	  logger.logValue ( Logger.SUPPORT,
			    Double.toString ( m_minSupport ) );

	} // end if we are logging
	
      } // end else use the cache file specified

      m_minSupport -= m_delta;

      m_minSupport = (m_minSupport < m_lowerBoundMinSupport) 
	? 0 
	: m_minSupport;
      necSupport = (int)(m_minSupport * (double)instances.numInstances() + 0.5);
      m_cycles++;
    }
    while ( ( m_rulesFound.size() < m_maxRulesOutput ) && 
	    ( m_minSupport >= m_lowerBoundMinSupport ) && 
	    ( m_minSupport * (double)instances.numInstances() + 0.5 >= 1 ) );
    {
      m_minSupport += m_delta;

    } // end while

  } // END public void mineAssociations ( Instances instances )
  // throws Exception 

//********************************************************//

  /**
   * Calculates the confidence of rules containing event
   * items in both the antecedent and consequent.
   *
   * @param rules
   * association rules already found
   *
   * @param instanceReader
   * data set from which to calculate confidence
   *
   * @param minConfidence
   * minimum confidence, rules which do not meet this are 
   * not included in the list of rules returned
   *
   * @return the new list of rules
   */

  private Vector eventConfidence ( Vector rules, 
				   DBReader instanceReader,
				   float minConfidence )
  {
    // move rules we have to check into seperate list
    Vector eventRuleVector = new Vector();

    int i;

    for ( i = 0; i < rules.size(); i++ )
    {
      if ( ((AssociationRule)rules.get ( i )).antecedentHasEvents() )
      {
	if ( ((AssociationRule)rules.get ( i )).consequentHasEvents() )
	{
	  eventRuleVector.add ( rules.get ( i ) );
	  rules.removeElementAt ( i );
	  
	  // don't forget to decrement i so we don't skip
	  // any rules
	  i--;

	} // end if consequent has events

      } // end if antecedent has events

    } // end for each rule in rule list

    // for each instance in the data set count how many times
    // the antecedent appears. Also count the number of times
    // the antecedent can be matched with a valid consequent.
    // the percentage of antecedents found and those matched
    // by a consequent will be the confidence of the rule.
    
    ARMinerItemset instance = null;

    try
    {
      instance = instanceReader.getFirstRow();
    } // end try
    catch ( Exception e )
    {
      System.err.println ( "AprioriSetsAndSequences.eventConfidence : " +
			   "Error scanning database!\n" + e );
    }

    // for each rule with events reset the event 
    // weights and weigh it against the first instance
    // of the data set

    // print out status, need the .0
    // so we get a double back...otherwise when num rows less
    // than 50 we get print status = 0 and that's no good -kap 2004-03-15

    int printStatus = (int)Math.ceil ( instanceReader.getNumRows() / 50.0 );

    System.out.println ( "AprioriSetsAndSequences.eventConfidence: " +
			 "scanning data set: " + instanceReader.getNumRows() + 
			 " rows, 1:" + printStatus + " - " );

    // temp rule
    AssociationRule tempRule;

    for ( i = 0; i < eventRuleVector.size(); i++ )
    {
      tempRule = (AssociationRule)eventRuleVector.get ( i );

      tempRule.resetEventWeights();
      tempRule.countEventWeights ( instance );

      switch ( debug ) {
      case 2: //case 0:
	System.out.println ( "AprioriSetsAndSequences.eventConfidence: " +
			     "Event weights for rule: " + tempRule +
			     " antecedent: " + 
			     tempRule.getAntecedentEventWeight() +
			     " rule: " + 
			     tempRule.getRuleEventWeight() );
      } // end switch debug

    } // end for each event rule

    // print out status of counting supports
    switch ( printStatus ) 
    {
    case 1: // only print a | if we're doing 1 for 1
      System.out.print ( "|" );

    } // end switch print status

    // count the rows
    int numRow = 1;

    // now we count the weights of the rules with the rest 
    // of the data set

    while ( instanceReader.hasMoreRows() )
    {
      try
      {
	instance = instanceReader.getNextRow();
      } // end try
      catch ( Exception e )
      {
      System.err.println ( "AprioriSetsAndSequences.eventConfidence : " +
			   "Error scanning database!\n" + e );
      }

      for ( i = 0; i < eventRuleVector.size(); i++ )
      {
	tempRule = (AssociationRule)eventRuleVector.get ( i );
	
	tempRule.countEventWeights ( instance );

	switch ( debug ) {
	case 2: //case 0:
	  System.out.println ( "AprioriSetsAndSequences.eventConfidence: " +
			       "Event weights for rule: " + tempRule +
			       " antecedent: " + 
			       tempRule.getAntecedentEventWeight() +
			       " rule: " + 
			       tempRule.getRuleEventWeight() );
	} // end switch debug
	
      } // end for each event rule

      // print out status 
      switch ( (int)( ( numRow + 1 ) % printStatus ) ) 
      {
      case 0:
	System.out.print ( "|" );
	
      } // end switch print status
      
      // increment row count
      numRow++;

    } // end while there are more instances in data set

    // now we must calculate the confidence of the rules
    // using event weight and add those that meet the
    // minimum to the rule to return

    for ( i = 0; i < eventRuleVector.size(); i++ )
    {
      tempRule = (AssociationRule)eventRuleVector.get ( i );

      if ( tempRule.eventWeightConfidence() >= minConfidence )
      {
	rules.add ( tempRule );

      } // end if confidence large enough

    } // end for each event rule

    return ( rules );

  } // END private Vector eventConfidence ( Vector rules, 
  //				   DBReader instanceReader,
  //				   float minConfidence )

//********************************************************//
  
  /**
   * Returns an enumeration describing the available options
   * 
   * @return an enumeration of all the available options
   */

  public Enumeration listOptions() 
  {
    String 
      maxRulesOutputHint = 
      "\tThe required number of rules. (default = " + m_maxRulesOutput + ")",
      minConfidenceHint = 
      "\tThe minimum confidence of a rule. (default = " 
      + m_minConfidence + ")",
      deltaHint = 
      "\tThe delta by which the minimum support is decreased in\n" +
      "\teach iteration. (default = " + m_delta + ")",
      lowerBoundMinSupportHint = 
      "\tThe lower bound for the minimum support. (default = " + 
      m_lowerBoundMinSupport + ")",
      upperBoundMinSupportHint = 
      "\tThe upper bound for the minimum support. (default = 1.0)",
      requiredAntecedentAttributesHint = 
      "\tThe set of attributes required to be in the antecedent.",
      requiredConsequentAttributesHint = 
      "\tThe set of attributes required to be in the consequent.",
      maxAntecedentsHint = 
      "\tThe maximum number of attributes in the antecedent. (0 = no maximum)",
      minAntecedentsHint = 
      "\tThe minimum number of attributes in the antecedent. (0 = no minimum)",
      maxConsequentsHint = 
      "\tThe maximum number of attributes in the consequent. (0 = no maximum)",
      minConsequentsHint = 
      "\tThe minimum number of attributes in the consequent. (0 = no minimum)",
      maxEventsHint = 
      "\tThe maximum number of event items of the same type allowed in a rule."
      + " (default = 2)",
      cacheFileNameHint = 
      "\tThe file from which to load frequent itemsets from",
      logStatsHint = 
      "\tSpecifies statistics should be written to log file",
      logStatsFileNameHint = 
      "\tSpecifies the log file statistics should be written to";
    
    FastVector newVector = new FastVector ( 12 );
    
    newVector.addElement 
      ( new Option ( maxRulesOutputHint, "N", 1, 
		     "-N <required number of rules output>" ) );
    newVector.addElement 
      ( new Option ( minConfidenceHint, "C", 1, 
		     "-C <minimum confidence score of a rule>" ) );
    newVector.addElement 
      ( new Option ( deltaHint, "D", 1,
		     "-D <delta for minimum support>" ) );
    newVector.addElement 
      ( new Option ( upperBoundMinSupportHint, "U", 1,
		     "-U <upper bound for minimum support>" ) );
    newVector.addElement 
      ( new Option ( lowerBoundMinSupportHint, "M", 1,
		     "-M <lower bound for minimum support>" ) );
    newVector.addElement 
      ( new Option ( requiredAntecedentAttributesHint, "A", 1, 
		     "-A <range of required antecedent attributes>" ) );
    newVector.addElement 
      ( new Option ( requiredConsequentAttributesHint, "Y", 1, 
		     "-Y <range of required consequent attributes>" ) );
    newVector.addElement 
      ( new Option ( maxAntecedentsHint, "B", 1, 
		     "-B <maximum antecedent items>" ) );
    newVector.addElement 
      ( new Option ( minAntecedentsHint, "B", 1, 
		     "-E <minimum antecedent items>" ) );
    newVector.addElement
      ( new Option ( maxEventsHint, "W", 1,
		     "-W <maximum events of same type allowed in a rule>" ) );
    newVector.addElement 
      ( new Option ( maxConsequentsHint, "X", 1, 
		     "-X <maximum consequent items>" ) );
    newVector.addElement 
      ( new Option ( minConsequentsHint, "Z", 1, 
		     "-Z <minimum consequent items>" ) );
    newVector.addElement 
      ( new Option ( cacheFileNameHint, "F", 1, 
		     "-F <cache file name>" ) );
    newVector.addElement 
      ( new Option ( logStatsHint, "L", 0, 
		     "-L" ) );
    newVector.addElement 
      ( new Option ( logStatsFileNameHint, "K", 1, 
		     "-K <log file name>" ) );

    return ( newVector.elements() );

  } // END public Enumeration listOptions() 

//********************************************************//
  
  /**
   * Parses a given list of options. Valid options are:<p>
   * 
   * -N required number of rules <br>
   * The required number of rules (default: 10). <p>
   * 
   * -C minimum confidence score of a rule <br>
   * The minimum confidence of a rule (default: 0.9). <p>
   * 
   * -D delta for minimum support <br>
   * The delta by which the minimum support is decreased in
   * each iteration (default: 0.05).
   * 
   * -U upper bound for minimum support <br>
   * The upper bound for minimum support. Don't explicitly look for 
   * rules with more than this level of support. <p>
   * 
   * -M lower bound for minimum support <br>
   * The lower bound for the minimum support (default = 0.1). <p>
   * 
   * -V <br>
   * If set then progress is reported iteratively during execution. <p>
   * 
   * -A <br>
   * The range of required attributes for the antecedents. <p>
   * 
   * -B <br>
   * The maximum number of antecedent items. <p>
   * 
   * -W <br>
   * Maximum number of event items of the same type allowed in a rule. <p>
   * 
   * -Y <br>
   * The range of required attributes for the consequents. <p>
   * 
   * -Z <br>
   * The minimum number of consequent items. <p>
   * 
   * -F <br>
   * The cache of frequent itemsets to use.
   *
   * -L <br>
   * Write statistics to a log file. <p>
   * 
   * @param options the list of options as an array of strings
   * @exception Exception if an option is not supported 
   */

  public void setOptions ( String[] options ) throws Exception
  {
    resetOptions();
    
    String numRulesString = Utils.getOption ( 'N', options ),

      minConfidenceString = Utils.getOption ( 'C', options ),
      deltaString = Utils.getOption ( 'D', options ),

      maxSupportString = Utils.getOption ( 'U', options ),
      minSupportString = Utils.getOption ( 'M', options ),

      requiredAntecedentsString = Utils.getOption ( 'A', options ),
      requiredConsequentsString = Utils.getOption ( 'Y', options ),

      minAntecedentString = Utils.getOption ( 'E', options ),
      maxAntecedentString = Utils.getOption ( 'B', options ),
      minConsequentString = Utils.getOption ( 'Z', options ),
      maxConsequentString = Utils.getOption ( 'X', options ),

      maxEventsString = Utils.getOption ( 'W', options ),
      cacheFileNameString = Utils.getOption ( 'F', options ),
      logStatsFileNameString = Utils.getOption ( 'K', options );

    boolean logStatsFlag = Utils.getFlag ( 'L', options );
    
    if ( numRulesString.length() != 0 )
    {
      m_maxRulesOutput = Integer.parseInt ( numRulesString );
    }
    if ( minConfidenceString.length() != 0 ) 
    {
      m_minConfidence = ( new Double ( minConfidenceString ) ).doubleValue();
    }
    if ( deltaString.length() != 0 ) 
    {
      m_delta = ( new Double ( deltaString ) ).doubleValue();
    }
    if ( maxSupportString.length() != 0 ) 
    {
      setUpperBoundMinSupport ( ( new Double ( maxSupportString ) ).
				doubleValue() );
    }
    if ( minSupportString.length() != 0 ) 
    {
      m_lowerBoundMinSupport = ( new Double ( minSupportString ) ).
	doubleValue();
    }
    if ( requiredAntecedentsString.length() != 0 )
    {
      setRequiredAntecedents ( requiredAntecedentsString );
    }
    if ( requiredConsequentsString.length() != 0 )
    {
      setRequiredConsequents ( requiredConsequentsString );
    }
    if ( minAntecedentString.length() != 0 )
    {
      m_minAntecedent = ( new Integer ( minAntecedentString ).intValue() );
    }
    if ( maxAntecedentString.length() != 0 )
    {
      m_maxAntecedent = ( new Integer ( maxAntecedentString ).intValue() );
    }
    if ( minConsequentString.length() != 0 )
    {
      m_minConsequent = ( new Integer ( minConsequentString ).intValue() );
    }
    if ( maxConsequentString.length() != 0 )
    {
      m_maxConsequent = ( new Integer ( maxConsequentString ).intValue() );
    }
    if ( maxEventsString.length() != 0 )
    {
      maxEvents = ( new Integer ( maxEventsString ).intValue() );
    }
    if ( cacheFileNameString.length() != 0 )
    {
      cacheFileName = cacheFileNameString;
    }
    if ( logStatsFlag )
    {
      // made the file name an option -kap -2004-01-05
      if ( logStatsFileNameString.length() != 0 )
      {
	logStatsFileName = logStatsFileNameString;
      }
      // have to use set method since it sets up the log
      // file -kap 2003-08-14
      setLogStats ( true );
    }
    else
    {
      if ( logStatsFileNameString.length() != 0 )
      {
	setLogStatsFileName ( logStatsFileNameString );
      }
    } // end else

  } // END public void setOptions ( String[] options ) throws Exception

//********************************************************//
  
  /**
   * Gets the current settings of the Apriori object.
   * 
   * @return an array of strings suitable for passing to setOptions
   */

  public String [] getOptions() 
  {
    String [] options = new String [ 30 ];
    int current = 0;
    
    options [ current++ ] = "-N"; 
    options [ current++ ] = "" + m_maxRulesOutput;

    options [ current++ ] = "-C"; 
    options [ current++ ] = "" + m_minConfidence;

    options [ current++ ] = "-D"; 
    options [ current++ ] = "" + m_delta;

    options [ current++ ] = "-U"; 
    options [ current++ ] = "" + m_upperBoundMinSupport;

    options [current++] = "-M"; 
    options [ current++ ] = "" + m_lowerBoundMinSupport;
    
    if ( getRequiredAntecedents() != "" )
    {
      options [ current++ ] = "-A"; 
      options [ current++ ] = getRequiredAntecedents();
    }
    if ( getRequiredConsequents() != "" )
    {
      options [ current++ ] = "-Y"; 
      options [ current++ ] = getRequiredConsequents();
    }
    
    options [ current++ ] = "-E"; 
    options [ current++] = "" + m_minAntecedent;
    
    options [ current++ ] = "-B"; 
    options [ current++] = "" + m_maxAntecedent;
    
    options [ current++ ] = "-Z"; 
    options [ current++ ] = "" + m_minConsequent;
    
    options [ current++ ] = "-X"; 
    options [ current++ ] = "" + m_maxConsequent;
    
    options [ current++ ] = "-W"; 
    options [ current++ ] = "" + maxEvents;
    
    if ( cacheFileName.length() != 0 )
    {
      options [ current++ ] = "-F"; 
      options [ current++ ] = "" + cacheFileName;
    }

    if ( logStatsFileName.length() != 0 )
    {
      options [ current++ ] = "-K";
      options [ current++ ] = "" + logStatsFileName;
    }

    if ( logStats == true )
    {
      options [ current++ ] = "-L";
    }

    while ( current < options.length ) 
    {
      options [ current++ ] = "";
    }
    
    return ( options );

  } // END public String [] getOptions() 

//********************************************************//
  
  /**
   * Outputs the association rules and their confidences
   * and supports.
   */

  public String toString() 
  {
    StringBuffer text = new StringBuffer();
    Enumeration e;
    AssociationRule as;
    int i;
    boolean comma = false;
    
    text.append ( "Required Attributes in Antecedents:\n" );
    
    for ( i = 0; i < m_instances.numAttributes(); i++ )
    {
      if ( m_requiredAntecedentAttributes.isInRange ( i ) )
      {
	if ( comma )
	{
	  text.append ( ", " );
	}
	else
	{
	  comma = true;
	}
	
	text.append ( m_instances.attribute ( i ).name() );
      }
    }
    
    if ( !comma )
    {
      text.append ( "none" );
    }
    
    comma = false;
    
    text.append ( "\n\nRequired Attributes in Consequents:\n" );
    
    for ( i = 0; i < m_instances.numAttributes(); i++ )
    {
      if ( m_requiredConsequentAttributes.isInRange ( i ) )
      {
	if ( comma )
	{
	  text.append ( ", " );
	}
	else
	{
	  comma = true;
	}
	
	text.append ( m_instances.attribute ( i ).name() );
      }
    }
    
    if ( ! comma )
    {
      text.append ( "none" );
    }
    
    text.append ( "\n\n" );
    
    e = m_rulesFound.elements();
    
    while ( e.hasMoreElements() )
    {
      as = (AssociationRule)e.nextElement();
      
      for ( i = 0; i < as.antecedentSize(); i++ )
      {
	if ( i > 0 )
	{
	  text.append ( " && " );
	}
	
	// We have to ask the itemset that I hacked in the association
	// rule class for the value of the item to support time sequence
	// attributes. -kap
	//	text.append ( (String)m_attributeHash.
	//	      get ( new Integer ( as.getAntecedentItem ( i ) ) ) );
	text.append ( as.originalItemset.
		      itemValue ( as.getAntecedentItem ( i ) ) );
      }
      
      text.append ( " ==> " );
      
      for ( i = 0; i < as.consequentSize(); i++ )
      {
	if ( i > 0 )
	{
	  text.append ( " && " );
	}
	
	// We have to ask the itemset that I hacked into the association
	// rule class for the value of the item to support time sequence
	// attributes. -kap
	//	text.append ( (String)m_attributeHash.
	//		      get ( new Integer ( as.getConsequentItem ( i ) ) ) );
	text.append ( as.originalItemset.
		      itemValue ( as.getConsequentItem ( i ) ) );

      }
      
      text.append ( "  [Conf: " + as.getConfidence() + 
		    ", Sup: " + as.getSupport() + 
		    ", Lift: " + as.getLift() +
		    ", Chi-S: " + as.getChiSquare() +
		    ", Event Weight: " +
		    as.getRuleEventWeight() + "]\n" );
    }
    
    return ( text.toString() );

  } // END public String toString() 

//********************************************************//
  
  public String requiredAntecedentsTipText()
  {
    return ( "Attributes required to be in the antecedent of the rules." );
  } // END public String requiredAntecedentsTipText()

//********************************************************//

  public void setRequiredAntecedents ( String s )
  {
    m_requiredAntecedentAttributes.setRanges ( s );
  } // END public void setRequiredAntecedents ( String s )

//********************************************************//

  public String getRequiredAntecedents()
  {
    return ( m_requiredAntecedentAttributes.getRanges() );
  } // END public String getRequiredAntecedents()

//********************************************************//

  public String requiredConsequentsTipText()
  {
    return ( "Attributes required to be in the consequent of the rules." );
  } // END public String requiredConsequentsTipText()

//********************************************************//

  public void setRequiredConsequents ( String s )
  {
    m_requiredConsequentAttributes.setRanges ( s );
  } // END public void setRequiredConsequents ( String s )

//********************************************************//

  public String getRequiredConsequents()
  {
    return ( m_requiredConsequentAttributes.getRanges() );
  } // END public String getRequiredConsequents()

//********************************************************//
  
  public String minAntecedentTipText()
  {
    return ( "Minimum number of items allowed in the " +
	     "antecedent of the rules." );
  } // END public String minAntecedentTipText()

//********************************************************//

  public void setMinAntecedent ( int i )
  {
    m_minAntecedent = i;
  } // END public void setMinAntecedent ( int i )

//********************************************************//

  public int getMinAntecedent()
  {
    return ( m_minAntecedent );
  } // END public int getMinAntecedent()

//********************************************************//
  
  public String maxAntecedentTipText()
  {
    return ( "Maximum number of items allowed in the " +
	     "antecedent of the rules." );
  } // END public String maxAntecedentTipText()

//********************************************************//

  public void setMaxAntecedent ( int i )
  {
    m_maxAntecedent = i;
  } // END public void setMaxAntecedent ( int i )

//********************************************************//

  public int getMaxAntecedent()
  {
    return ( m_maxAntecedent );
  } // END public int getMaxAntecedent()

//********************************************************//

  public String minConsequentTipText()
  {
    return ( "Minimum number of items allowed in the " +
	     "consequent of the rules." );
  } // END public String minConsequentTipText()

//********************************************************//

  public void setMinConsequent ( int i )
  {
    m_minConsequent = i;
  } // END public void setMinConsequent ( int i )

//********************************************************//

  public int getMinConsequent()
  {
    return ( m_minConsequent );
  } // END public int getMinConsequent()

//********************************************************//

  public String maxConsequentTipText()
  {
    return ( "Maximum number of items allowed in the " +
	     "consequent of the rules." );
  } // END public String maxConsequentTipText()

//********************************************************//

  public void setMaxConsequent ( int i )
  {
    m_maxConsequent = i;
  } // END public void setMaxConsequent ( int i )

//********************************************************//

  public int getMaxConsequent()
  {
    return ( m_maxConsequent );
  } // END public int getMaxConsequent()

//********************************************************//

  public String maxEventsTipText()
  {
    return ( "Maximum number of event items of the same type " +
	     "allowed in a rule." );
  } // END public String maxEventsTipText()

//********************************************************//

  public void setMaxEvents ( int i )
  {
    maxEvents = i;
  } // END public void setMaxEvents ( int i )

//********************************************************//

  public int getMaxEvents()
  {
    return ( maxEvents );
  } // END public int getMaxEvents()

//********************************************************//

  public String upperBoundMinSupportTipText() 
  {
    return ( "Upper bound for minimum support. Start iteratively decreasing "
      +"minimum support from this value." );
  } // END public String upperBoundMinSupportTipText() 

//********************************************************//

  public double getUpperBoundMinSupport()
  {
    return ( m_upperBoundMinSupport );
  } // END public double getUpperBoundMinSupport()

//********************************************************//

  public void setUpperBoundMinSupport ( double v ) 
  {
    m_upperBoundMinSupport = v;
  } // END public void setUpperBoundMinSupport ( double v ) 

//********************************************************//

  public String lowerBoundMinSupportTipText() 
  {
    return ( "Lower bound for minimum support." );
  } // END public String lowerBoundMinSupportTipText() 

//********************************************************//

  public double getLowerBoundMinSupport() 
  {
    return ( m_lowerBoundMinSupport );
  } // END public double getLowerBoundMinSupport() 

//********************************************************//

  public void setLowerBoundMinSupport ( double v ) 
  {
    m_lowerBoundMinSupport = v;
  } // END public void setLowerBoundMinSupport ( double v ) 

//********************************************************//

/* Changed from minMetric to minConfidence so people know what
   this actually is. -kap */

  public String minConfidenceTipText() 
  {
    return ( "Minimum confidence score. " +
	     "Consider only rules with scores higher than " +
	     "this value." );
  } // END public String minMetricTipText() 

//********************************************************//

/* Changed from minMetric to minConfidence so people know what
   this actually is. -kap */

  public double getMinConfidence() 
  {
    return ( m_minConfidence );
  } // END public double getMinMetric() 

//********************************************************//

/* Changed from minMetric to minConfidence so people know what
   this actually is. -kap */

  public void setMinConfidence ( double v ) 
  {
    m_minConfidence = v;
  } // END public void setMinMetric ( double v ) 

//********************************************************//

  public String numRulesTipText() 
  {
    return ( "Number of rules to find." );
  } // END public String numRulesTipText() 

//********************************************************//

  public int getNumRules() 
  {
    return ( m_maxRulesOutput );
  } // END public int getNumRules() 

//********************************************************//

  public void setNumRules ( int v ) 
  {
    m_maxRulesOutput = v;
  } // END public void setNumRules ( int v ) 

//********************************************************//

  public String deltaTipText() 
  {
    return ( "Iteratively decrease support by this factor. Reduces support " +
	     "until min support is reached or required " +
	     "number of rules has been generated." );
  } // END public String deltaTipText() 

//********************************************************//

  public double getDelta() 
  {
    return ( m_delta );
  } // END public double getDelta() 

//********************************************************//

  public void setDelta ( double v ) 
  {
    m_delta = v;
  } // END public void setDelta ( double v ) 

//********************************************************//

  public String cacheFileNameTipText() 
  {
    return ( "Frequent itemsets are loaded from this cache file. " +
	     "If specified rules will be built directly from these " +
	     "itemsets rather than mining them from the data set. " +
	     "Please note that the cache should be the result of " +
	     "mining frequent itemsets from the data set specified." );
  } // END public String cacheFileNameTipText() 

//********************************************************//

  public String getCacheFileName() 
  {
    return ( cacheFileName );
  } // END public String getCacheFileName() 

//********************************************************//

  public void setCacheFileName ( String v ) 
  {
    cacheFileName = v;
  } // END public void setCacheFileName ( String v ) 

//********************************************************//

  public void setLogStatsFileName ( String v ) 
  {
    logStatsFileName = v;

    // make sure to turn on logging
    // assume that if a log name if specified that the user
    // wants info logged.
    setLogStats ( true );

  } // END public void setCacheFileName ( String v ) 

//********************************************************//

  public String logStatsFileNameTipText()
  {
    return ( "Set this option to specify the log file " +
	     "statistics are written to. The log file will " + 
	     "be named " + Logger.defaultLogStatsFileName +
	     " if this option is not set." );
  } // END public String logStatsFileNameTipText()

//********************************************************//

  public String getLogStatsFileName()
  {
    return ( logStatsFileName );
  } // END public boolean getLogStatsFileName()

//********************************************************//

  public String logStatsTipText()
  {
    return ( "Set this option to have statistics written " +
	     "to a log file. The log file will be named " +
	     Logger.defaultLogStatsFileName +
	     ".  Statistics are added to the " +
	     "log if one already exists." );
  } // END public String logStatsTipText()

//********************************************************//

  public boolean getLogStats()
  {
    return ( logStats );
  } // END public boolean getLogStats()

//********************************************************//

  public void setLogStats ( boolean l )
  {
    logStats = l;

    if ( logStats )
    {
      // are we specifying a file name?
      if ( logStatsFileName.length() == 0 )
      {
	logStatsFileName = Logger.defaultLogStatsFileName;

      } // end if specifying log file name
      
    } // end if we are logging stats

  } // END public void setLogStats ( boolean l )

//********************************************************//
  
  /**
   * Main method for testing this class.
   */

  public static void main ( String[] options ) 
  {
    String trainFileString;
    StringBuffer text = new StringBuffer();
    AprioriSetsAndSequences ass = new AprioriSetsAndSequences();
    Reader reader;

    try 
    {
      text.append ( "\n\nApriori Sets And Sequences options:\n\n" );
      text.append ( "-t <training file>\n" );
      text.append ( "\tThe name of the training file.\n" );

      Enumeration enum = ass.listOptions();

      while ( enum.hasMoreElements() )
      {
	Option option = (Option)enum.nextElement();
	text.append ( option.synopsis() + '\n' );
	text.append ( option.description() + '\n' );

      } // end while more options

      trainFileString = Utils.getOption ( 't', options );

      if ( trainFileString.length() == 0 ) 
      {
	throw new Exception ( "No training file given. " +
			      "Please specify one using the -t option." );
      }
      
      ass.setOptions ( options );

      reader = new BufferedReader ( new FileReader ( trainFileString ) );

      ass.buildAssociations ( new Instances ( reader ) );

      System.out.println ( ass );

    } // end try
    catch ( Exception e ) 
    {
      e.printStackTrace();
      System.out.println ( "\n" + e.getMessage() + text );

    } //end catch

  } // END public static void main ( String[] options ) 

//********************************************************//

} // END public class AprioriSetsAndSequences extends Associator 
// implements OptionHandler 
