/*
ARMiner - Association Rules Miner
Copyright (C) 2000  UMass/Boston - Computer Science Department

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or (at
your option) any later version.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
USA

The ARMiner Server was written by Dana Cristofor and Laurentiu
Cristofor.

The ARMiner Client was written by Abdelmajid Karatihy, Xiaoyong Kuang,
and Lung-Tsung Li.

The ARMiner package is currently maintained by Laurentiu Cristofor
(laur@cs.umb.edu).
*/

package wpi.associations.arminerSequence;

import java.util.Arrays;
import java.util.Vector;

/*
   This file is a part of the ARMiner project.
   
   (P)1999-2000 by ARMiner Server Team:

   Dana Cristofor
   Laurentiu Cristofor

   modified by Keith A. Pray ( kap@wpi.edu )
*/

/*
  Maintenance log started on April 29th, 2002

  April   29, 2002   - Clarified exception statements -kap

  May     26, 2002   - Added time sequence support
                       Cleaned up formatting -kap

  February 9, 2003   - Added proper event item comparison
                       to the equals method -kap

  February 24, 2003  - made consistent the convention of using a lower case
		       letter to start a method name. -kap

  August 13, 2003    - added antecedentSupport, consequentSupport,
                       and numInstances data members. -kap
		     - added getAntecedentSupport, getConsequentSupport,
                       getChiSquare, and getLift methods. -kap

  September 25, 2003 - added antecedentEventWeight and consequentEventWeight
                       data members. needed consequent event weight for 
		       calculating lift for rules with events. added 
		       antecedent event weight while I was at it. -kap

  September 27, 2003 - modified Chi-Squared to result in 0 if support and
                       confidence are equal or if lift and confidence are 
		       equal. Either of these conditions would result
		       in a divide by 0 error. -kap

  December 2, 2003   - added ruleEventWeight data member. this is
                       needed to support an antecedent context based 
		       confidence of the association rule. -kap

  April 29, 2004     - fixed countEventWeights to check for regular items 
                       too -kap
*/

/**
   AssociationRule.java<P>

   An association rule has two parts: the antecedent of
   the rule and the consequent of the rule, both of which
   are sets of items.
   Associated with these are a support and a confidence. 
   The support tells how many rows of a database support
   this rule, the confidence tells what percentage of the
   rows that contain the antecedent also contain the
   consequent.

@author Dana Cristofor
@author Laurentiu Cristofor
@author Keith A. Pray (kap@wpi.edu)
*/

public class AssociationRule 
  implements java.io.Serializable
{

//********************************************************//
// ----- Data Members -----
//********************************************************//  

  public static final int ANTECEDENT_SIZE = 1;
  public static final int CONSEQUENT_SIZE = 2;
  public static final int SUPPORT         = 3;
  public static final int CONFIDENCE      = 4;

  /** The antecedent. @serial */
  private int[] antecedent;

  /** The consequent. @serial */
  private int[] consequent;

  /** The support of the association rule. @serial */
  private float support;

  /** The support of the antecedent of association rule. @serial */
  private float antecedentSupport;

  /** The support of the consequent of association rule. @serial */
  private float consequentSupport;

  /** The event weight of the antecedent of association rule. @serial */
  private float antecedentEventWeight;

  /** The event weight of the consequent of association rule. @serial */
  private float consequentEventWeight;

  /** The event weight of the association rule. This is the 
   * number of times the consequent is found to match a 
   * previously found antecedent.  @serial */
  private float ruleEventWeight;

  /** The confidence of the association rule. @serial */
  private float confidence;

  /** Number of instances in the data set. Used for computing statistical
      measures */
  private int numInstances;

  /** the original item set the rule's antecedent and
      consequent were derived from */
  public ARMinerItemset originalItemset;

  /** Specifies debug info level 
   * 0: no debug info
   * 1: input to methods
   * 2: and output from methods
   * 3: and all sorts of stuff 
   */
  private static final int debug = 0;

//********************************************************//
// ----- Methods-----
//********************************************************//

  /**
   * Sets the number of instances contained in the data set
   * from which the frequent item set were mined.
   *
   * @param n
   * the number of instances or rows
   */

  public void setNumInstances ( int n )
  {
    numInstances = n;

  } // END public void setNumInstances ( int n )

//********************************************************//

  /**
   * Returns the number of instances contained in the data set
   * from which the frequent item set were mined.
   *
   * @return the number of instances or rows
   */

  public int getNumInstances()
  {
    return ( numInstances );

  } // END public int getNumInstances()

//********************************************************//

  /* This method was changed to include an extra parameter,
     the original item set from which the antecedent and
     consequent were derived. This is needed to get the
     value of the events contained in the rule. -kap */

  /**
   * Creates a new association rule.
   *
   * @param antecedent   
   * the antecedent of the association rule
   * 
   * @param consequent   
   * the consequent of the association rule
   *
   * @param original
   * the original item set the antecedent and consequent
   * were derived from
   *
   * @param support  
   * the support of the association rule
   *
   * @param confidence  
   * the confidence of the association rule
   *
   * @exception IllegalArgumentException  
   * <code>antecedent</code> 
   * or <code>consequent</code> are null or <code>support</code>
   * or <code>confidence</code> are not between 0 and 1
   */

  public AssociationRule ( ARMinerItemset antecedent, 
			   ARMinerItemset consequent,
			   ARMinerItemset original,
			   float support, float confidence )
  {
    // the param checking below used to be all in one and the
    // error message was very cryptic. Hopefully this will be a bit
    // more useful. -kap
    if ( antecedent == null )
    {
      throw new IllegalArgumentException
	( "AssociationRule: constructor: antecedent null." );
    }
    if ( consequent == null )
    {
      throw new IllegalArgumentException
	( "AssociationRule: constructor: consequent null." );
    }
    if ( original == null )
    {
      throw new IllegalArgumentException
	( "AssociationRule: constructor: original null." );
    }
    if ( ( support < 0 ) || support > 1 )
    {
      throw new IllegalArgumentException
	( "AssociationRule: constructor: support = " + support + ", " +
	  "must be between 0 and 1" );
    }
    if ( confidence < 0 )
    {
      throw new IllegalArgumentException
	( "AssociationRule: constructor: confidence = " + confidence + ", " +
	  "must be between 0 and 1" );

    } // end if confidence below min
    if  ( confidence > 1 )
    {
      System.err.println ( "AssociationRule constructor: " +
			   "should fail but I'm letting it slide since this " +
			   "is a bug in ARMiner - antecedent = " + antecedent +
			   " - consequent = " + consequent + 
			   " - support = " + support + 
			   " - confidence = " + confidence );

/* Commenting out until we find out why supports are acting funny

   Basic reason is that sometimes an item is written to cache but
   is not read back. -kap

      throw new IllegalArgumentException
	( "AssociationRule: constructor: confidence = " + confidence + ", " +
	  "must be between 0 and 1" );
*/
    } // end if confidence above max

    this.antecedent = new int [ antecedent.size() ];

    for ( int i = 0; i < antecedent.size(); i++ )
    {
      this.antecedent [ i ] = antecedent.getItem ( i );
    }

    this.consequent = new int [ consequent.size() ];

    for ( int i = 0; i < consequent.size(); i++ )
    {
      this.consequent [ i ] = consequent.getItem ( i );
    }

    this.support = support;
    this.antecedentSupport = antecedent.getSupport();
    this.consequentSupport = consequent.getSupport();

    this.antecedentEventWeight = antecedent.getEventWeight();
    this.consequentEventWeight = consequent.getEventWeight();

    this.confidence = confidence;

    // remember from where we came
    originalItemset = original;

  } // END public AssociationRule ( ARMinerItemset antecedent, 
  //			   ARMinerItemset consequent,
  //			   ARMinerItemset original,
  //			   float support, float confidence )

//********************************************************//

  /**
   * Return size of antecedent.
   *
   * @return size of antecedent
   */

  public int antecedentSize()
  {
    return ( antecedent.length );

  } // END public int antecedentSize()

//********************************************************//

  /**
   * Return size of consequent.
   *
   * @return size of consequent
   */

  public int consequentSize()
  {
    return ( consequent.length );

  } // END public int consequentSize()

//********************************************************//

  /**
   * Return support of association rule.
   */

  public float getSupport()
  {
    return ( support );

  } // END public float getSupport()

//********************************************************//

  /**
   * Return support of antecedent of association rule.
   */

  public float getAntecedentSupport()
  {
    return ( antecedentSupport );

  } // END public float getAntecedentSupport()

//********************************************************//

  /**
   * Return support of antecedent of association rule.
   */

  public float getConsequentSupport()
  {
    return ( consequentSupport );

  } // END public float getConsequentSupport()

//********************************************************//

  /**
   * Return event weight of antecedent of association rule.
   */

  public float getAntecedentEventWeight()
  {
    return ( antecedentEventWeight );

  } // END public float getAntecedentEventWeight()

//********************************************************//

  /**
   * Return event weight of consequent of association rule.
   */

  public float getConsequentEventWeight()
  {
    return ( consequentEventWeight );

  } // END public float getConsequentEventWeight()

//********************************************************//

  /**
   * Return event weight of association rule.
   */

  public float getRuleEventWeight()
  {
    return ( ruleEventWeight );

  } // END public float getRuleEventWeight()

//********************************************************//

  /**
   * Return confidence of association rule.
   */

  public float getConfidence()
  {
    return ( confidence );

  } // END public float getConfidence()

//********************************************************//

  /**
   * Calculates and records confidence of association rule
   * based on antecedent event weight and overall
   * association rule event weight in context of the
   * antecedent.
   *
   * @return the confidence value 
   */

  public float eventWeightConfidence()
  {
    confidence = ruleEventWeight / antecedentEventWeight;

    return ( confidence );

  } // END public float eventWeightConfidence()

//********************************************************//

  /**
   * Return lift of association rule.
   */

  public float getLift()
  {

    // if both antecedent and consequent have events
    // we'll use event weight instead of support
    if ( antecedentHasEvents() && consequentHasEvents() )
    {
      return ( originalItemset.getEventWeight() / consequentEventWeight );

    } // end if both have events

    return ( confidence / consequentSupport );

  } // END public float getLift()

//********************************************************//

  /**
   * Calculates and returns the Chi-square measure of the
   * rule.
   * <br>
   * Sergio A. Alvarez.
   * "Chi--Squared Computation for Association Rules: Preliminary Results''.
   * Technical Report BC-CS-2003-01, Computer Science Department,
   * Boston College, July 2003.
   * http://cs.bc.edu/research/reports/Alvarez:Chi2TR.html
   *
   * @return Chi-square measure of rule
   */

  public float getChiSquare()
  {
    // ok, we are going to use the formula described in the
    // reference cited above

    //                                  supp conf
    //  X^2 = n (lift - 1)^2 ( --------------------------- )
    //                         (conf - supp) (lift - conf)

    // if conf = supp or lift = conf then X^2 = 0

    if ( support == confidence )
    {
      return ( 0 );
    } // end if support = confidence

    float lift = getLift();

    if ( lift == confidence )
    {
      return ( 0 );
    } // end if lift = confidence

    float left = ( lift - 1 ) * ( lift - 1 ) * numInstances;

    float right = ( support * confidence ) / 
      ( ( confidence - support ) * ( lift - confidence ) );

    return ( left * right );

  } // END public float getChiSquare ( int numInstances )

//********************************************************//

  /**
   * Checks if there is at least one event in the
   * consequent.
   *
   * @return true if there is at least one event,
   * false otherwise
   */

  public boolean consequentHasEvents()
  {
    for ( int i = 0; i < consequent.length; i++)
    {
      if ( originalItemset.eventName ( consequent [ i ] ) != null )
      {
	return ( true );
      } // end if item i is an event

    } // end for each consequent item
    
    // I guess wee didn't find an event
    return ( false );

  } // END public boolean consequentHasEvents()

//********************************************************//

  /**
   * Checks if there is at least one event in the
   * antecedent.
   *
   * @return true if there is at least one event,
   * false otherwise
   */

  public boolean antecedentHasEvents()
  {
    for ( int i = 0; i < antecedent.length; i++)
    {
      if ( originalItemset.eventName ( antecedent [ i ] ) != null )
      {
	return ( true );
      } // end if item i is an event

    } // end for each consequent item
    
    // I guess wee didn't find an event
    return ( false );

  } // END public boolean antecedentHasEvents()

//********************************************************//

  /**
   * Return i-th item in antecedent.
   *
   * @param i   
   * the index of the item to get
   *
   * @exception IndexOutOfBoundsException   
   * <code>i</code> is an invalid index
   *
   * @return the <code>i</code>-th item in antecedent
   */

  public int getAntecedentItem ( int i )
  {
    if ( i < 0 || i >= antecedent.length )
    {
      throw new IndexOutOfBoundsException
	( "AssociationRule.getAntecedentItem ( i ) invalid index" );
    }

    return ( antecedent [ i ] );

  } // END public int getAntecedentItem ( int i )

//********************************************************//
  
  /**
   * Return i-th item in consequent.
   *
   * @param i   the index of the item to get
   * @exception IndexOutOfBoundsException   <code>i</code> is an invalid index
   * @return   the <code>i</code>-th item in consequent
   */

  public int getConsequentItem ( int i )
  {
    if (i < 0 || i >= consequent.length )
    {
      throw new IndexOutOfBoundsException
	( "AssociationRule.getConsequentItem ( i ) invalid index" );
    }

    return ( consequent [ i ] );

  } // END public int getConsequentItem ( int i )

//********************************************************//

  /**
   * Compare two AssociationRule objects on one of several
   * criteria.
   *
   * @param ar   
   * the AssociationRule object with which we want to
   * compare this object
   *
   * @param criteria   
   * the criteria on which we want to compare, can 
   * be one of ANTECEDENT_SIZE, CONSEQUENT_SIZE, SUPPORT
   * or CONFIDENCE.
   *
   * @exception IllegalArgumentException   
   * <code>ar</code> is null or criteria is invalid
   *
   * @return   
   * a negative value if this object is smaller than 
   * <code>ar</code>, 0 if they are equal, and a positive
   * value if this object is greater.
   */

  public int compareTo ( AssociationRule ar, int criteria )
  {
    if ( ar == null )
    {
      throw new IllegalArgumentException
	( "AssociationRule.compareTo ( ar, criteria ) ar null" );
    }

    float diff;

    if ( criteria == ANTECEDENT_SIZE )
    {
      return ( this.antecedent.length - ar.antecedent.length );
    }
    else
    {
      if ( criteria == CONSEQUENT_SIZE )
      {
	return ( this.consequent.length - ar.consequent.length );
      }
      else
      { 
	if ( criteria == SUPPORT )
	{
	  diff = this.support - ar.support;
	}
	else 
	{
	  if ( criteria == CONFIDENCE )
	  {
	    diff = this.confidence - ar.confidence;
	  }
	  else
	  {
	    throw new IllegalArgumentException
	      ( "AssociationRule.compareTo ( ar, criteria ) " +
		"invalid criteria" );
	  }
	}
      }
    }

    if ( diff < 0 )
    {
      return ( -1 );
    }
    else
    {
      if ( diff > 0 )
      {
	return ( 1 );
      }
      else 
      {
	return ( 0 );
      }
    }

  } // END public int compareTo ( AssociationRule ar,
  // int criteria )

//********************************************************//

  /* Added further test to make sure event items are
     compared properly -kap */

  /**
   * Compare two AssociationRule objects on one of several
   * criteria.
   *
   * @param obj
   * the object to compare for equality
   *
   * @return
   * true if the object is equal
   */

  public boolean equals ( Object obj )
  {
    if ( ! ( obj instanceof AssociationRule ) || 
	 obj == null )
    {
      return ( false );
    } // end if object is not an association rule

    AssociationRule other = (AssociationRule)obj;

    if ( antecedent.length != other.antecedent.length )
    {
      return ( false );
    } // end if rule's antecedent is not the same length

    if ( consequent.length != other.consequent.length )
    {
      return ( false );
    } // end if rule's consequent is not the same length

    // have to compare the value of each item in the antecedents
    // and the consequences -kap

    // now, since the items may not be in the same order
    // we have to look for each item at each spot in the other
    // antecedent, let's track which are used to save time
    boolean [] used = new boolean [ antecedent.length ];
    
    // initialize used to false
    Arrays.fill ( used, false );

    for ( int i = 0; i < antecedent.length; i++ )
    {
      int j = 0;

      for ( j = 0; j < antecedent.length; j++ )
      {
	// only try this j if it isn't used
	if ( used [ j ] )
	{
	  // continue looking with the next j then
	  continue;
	} // end if this j was used
	
	if ( originalItemset.itemValue ( antecedent [ i ] ).equals
	     ( other.originalItemset.itemValue ( other.antecedent [ j ] ) ) )
	{
	  // mark this j as being used
	  used [ j ] = true;
	  // then we simply stop looking
	  break;
	  
	} // end if we found a match
	
      } // for each other antecedent item
      
      // if we didn't find the item we should return false
      if ( j >= antecedent.length )
      {
	return ( false );
      } // end if we went through all of other's items and
	// didn't find a match
      
    } // end for each antecedent item find a match in other

    // have to compare the value of each item in the consequence

    // have to make a new used array
    used = new boolean [ consequent.length ];

    // initialize used to false again
    Arrays.fill ( used, false );

    for ( int i = 0; i < consequent.length; i++ )
    {
      int j = 0;

      for ( j = 0; j < consequent.length; j++ )
      {
	// only try this j if it isn't used
	if ( used [ j ] )
	{
	  // continue looking with the next j then
	  continue;
	} // end if this j was used
	
	if ( originalItemset.itemValue ( consequent [ i ] ).equals
	     ( other.originalItemset.itemValue ( other.consequent [ j ] ) ) )
	{
	  // mark this j as being used
	  used [ j ] = true;
	  // then we simply stop looking
	  break;
	  
	} // end if we found a match
	
      } // for each other consequent item
      
      // if we didn't find the item we should return false
      if ( j >= consequent.length )
      {
	return ( false );
      } // end if we went through all of other's items and
	// didn't find a match
      
    } // end for each consequent item find a match in other

    return ( true );

  } // END public boolean equals ( Object obj )

//********************************************************//

  /**
   * Counts the number of occurrences of the antecedent
   * and the number of occurrences of the consequent in
   * context of the antecedents found in the specified
   * item set. These counts are stored in this association 
   * rule. To reset the count use resetEventWeights().
   * The counts are accumulative for each call of this 
   * method.
   *
   * @param itemset
   * the item set to find occurrences of the antecedent and 
   * consequent in
   *
   * @return an array of size 2, the first index holding
   * the count of antecedents, the second holding the count
   * of antecedents which had matching consequents.
   */

  public int[] countEventWeights ( ARMinerItemset itemset )
  {
    // the array of counts to return
    int counts[] = new int [ 2 ];

    // have to look up how to do this nicely -kap

    // initialize the counts to zero
    counts [ 0 ] = 0;
    counts [ 1 ] = 0;

    // first make sure there are events in both our 
    // antecedent and consequent. If this is not the case
    // we do not have to count event weight.
    if ( ! antecedentHasEvents() || ! consequentHasEvents() )
    {
      return ( counts );

    } // end if antecedent or consequent does not have events

    // build an item set representing the antecedent
    ARMinerItemset antecedentItemset = 
      ARMinerItemset.buildItemsetFromItemList ( originalItemset, antecedent );

    // first we must make sure that the antecedent is included in 
    // the itemset -kap 2004-04-29
    if ( ! antecedentItemset.isIncludedIn ( itemset ) )
    {
      // then there is no point to go further
      return ( counts );
    }

    // get possible maps of the given item set to the antecedent
    Vector antecedentMapVector = 
      antecedentItemset.findAllMaps ( itemset );

    // if we did not get any maps we can quit now
    if ( antecedentMapVector.size() < 1 )
    {
      return ( counts );

    } // end no maps found
    else
    {
      // record the number of maps found so we can return the value
      counts [ 0 ] = antecedentMapVector.size();

      // add the number of maps to the antecedent event weight
      antecedentEventWeight = antecedentEventWeight + counts [ 0 ];

    } // end else maps were found

    // before we go further and try to extend the antecedents 
    // to the consequent, let;s make sure the whole rule is
    // included in the item set -kap 2004-04-29
    if ( ! originalItemset.isIncludedIn ( itemset ) )
    {
      // no point in going further
      return ( counts );
    }
    
    // take all the maps we found for the antecedent and see
    // how many can be extended into maps for the entire original
    // itemset.
      
    // record the number of maps found so we can return the value
    counts [ 1 ] = 
      originalItemset.countMapExtensions ( itemset, antecedentMapVector );

    // add the number of maps to the antecedent event weight
    ruleEventWeight = ruleEventWeight + counts [ 1 ];

    // all done
    return ( counts );

  } // END public null countEventWeights ( ARMinerItemset itemset )

//********************************************************//

  /**
   * Resets the event weight count to zero. Use this when
   * beginning to count the event weights of this 
   * association rule. This might be necessary to accurately
   * calculate the confidence and lift of this rule.
   */

  public void resetEventWeights()
  {
    antecedentEventWeight = 0;
    consequentEventWeight = 0;
    ruleEventWeight = 0;

  } // END public void resetEventWeights()

//********************************************************//

  /**
   * Return a String representation of the AssociationRule.
   *
   * @return String representation of AssociationRule
   */

  public String toString()
  {
    String s = "{";

    for ( int i = 0; i < antecedent.length; i++ )
    {
      s += antecedent [ i ] + " ";
    }

    s += "}->{";

    for ( int i = 0; i < consequent.length; i++ )
    {
      s += consequent[i] + " ";
    }

    s += "} (s=" + support + ", c=" + confidence + 
      ", l= " + getLift() + ", cs=" + getChiSquare() + ")";

    return ( s );

  } // END public String toString()

//********************************************************//

  /** 
   * Main method for testing purposes only.
   */

  public static void main ( String[] args )
  {
    ARMinerItemset is1 = new ARMinerItemset();
    ARMinerItemset is2 = new ARMinerItemset();

    is1.addItem ( 7 );
    is1.addItem ( 3 );
    is1.addItem ( 15 );

    System.out.println ( "is1: " + is1 );

    is2.addItem ( 12 );
    is2.addItem ( 5 );
    is2.addItem ( 8 );

    System.out.println ( "is2: " + is2 );

    AssociationRule ar = new AssociationRule ( is1, is2, 
					       null,
					       (float)0.5055, 
					       (float)0.3033 ); 

    System.out.println ( "ar: " + ar );

  } // END public static void main ( String[] args )

//********************************************************//

} // END public class AssociationRule implements java.io.Serializable
