/*
ARMiner - Association Rules Miner
Copyright (C) 2000  UMass/Boston - Computer Science Department

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or (at
your option) any later version.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
USA

The ARMiner Server was written by Dana Cristofor and Laurentiu
Cristofor.

The ARMiner Client was written by Abdelmajid Karatihy, Xiaoyong Kuang,
and Lung-Tsung Li.

The ARMiner package is currently maintained by Laurentiu Cristofor
(laur@cs.umb.edu).
*/

/*
   This file is a part of the ARMiner project.
   
   (P)1999-2000 by ARMiner Server Team:

   Dana Cristofor
   Laurentiu Cristofor

   modified by Keith A. Pray 2002
*/

package wpi.associations.arminerSequence;

import java.util.Vector;
import weka.core.Utils;
import java.io.*;

/*
  Maintainance log started on June 11th, 2002

  June 11th, 2002    - changed main method to take a file name as 
                       an argument
		     - cleaned up formatting -kap

  July 6th, 2002     - added utility for comparing two cache files -kap

  February 24, 2003  - made consistent the convention of using a lower case
		       letter to start a method name. -kap
*/

/**
 * DBCacheReader.java
 *
 * A DBCacheReader deserializes itemsets from cache.
 *
 * @author Dana Cristofor
 * @author Laurentiu Cristofor
 * @author Keith A. Pray (kap@wpi.edu)
*/

public class DBCacheReader
{

//********************************************************//
// ----- Data Members -----
//********************************************************//  

  private ObjectInputStream instream ;
  private String filename;

//********************************************************//
// ----- Methods-----
//********************************************************//

  /**
   * Initializes a DBCacheReader to read from the specified cache file.
   *
   * @param name   name of the cache file
   * @exception IllegalArgumentException   <code>name</code> is null
   * @exception IOException   from java.io package
   */                                

  public DBCacheReader ( String name ) throws IOException
  {
    if ( name == null )
    {
      throw new 
	IllegalArgumentException ( "Constructor argument must be non null");
    }

    filename = name;
    instream = new ObjectInputStream ( new FileInputStream ( filename ) );

  } // END public DBCacheReader ( String name ) throws IOException

//********************************************************//

  /**
   * Closes the cache file.
   *
   * @exception IOException   from java.io package
   */
                                
  public void close() throws IOException
  {
    instream.close();

  } // END public void close() throws IOException

//********************************************************//

  /**
   * Return the first itemset from cache.
   *
   * @exception IOException
   * from java.io package
   *
   * @exception ClassNotFoundException
   * from java.io package
   *
   * @return first itemset in cache
   */

  public ARMinerItemset getFirstItemset()
    throws IOException, ClassNotFoundException
  {
    instream.close();
    instream = new ObjectInputStream ( new FileInputStream ( filename ) );
    
    return ( (ARMinerItemset)instream.readObject() );

  } // END public ARMinerItemset getFirstItemset()
  //    throws IOException, ClassNotFoundException

//********************************************************//

  /**
   * Return next itemset from cache.
   *
   * @exception IOException   from java.io package
   * @exception ClassNotFoundException   from java.io package
   * @return   next itemset in cache
   */
                           
  public ARMinerItemset getNextItemset()
    throws IOException, ClassNotFoundException
  {
    return ( (ARMinerItemset)instream.readObject() );

  } // END public ARMinerItemset getNextItemset()
  //    throws IOException, ClassNotFoundException

//********************************************************//

  /**
   * Returns the itemsets contained in this cache that
   * do not appear in the specified db cache reader.
   *
   * @param dbCacheReader
   * the cache to compare to
   */

  public Vector compare ( DBCacheReader dbCacheReader )
    throws IOException, ClassNotFoundException
  {
    ARMinerItemset ourItemset, theirItemset;
    Vector notFound = new Vector();

    // a vector containing itemsets from dbCacheReader not matched yet
    Vector theirNotFound = new Vector();
    int i;

    try 
    {
      ourItemset = this.getFirstItemset();
      theirItemset = dbCacheReader.getFirstItemset();

      while ( true )
      {
	try 
	{
	  // do we have our itemset in the their not found?
	  for ( i = theirNotFound.size() - 1; i > -1; i-- )
	  {
	    if ( ((ARMinerItemset)theirNotFound.get ( i ))
		 .isEqualTo ( ourItemset ) )
	    {
	      // remove it from their not found
	      theirNotFound.removeElementAt ( i );
	      break;

	    } // end if our itemset in their not found
	  } // end for each itemset in their not found

	  // did we find it?
	  if ( i < 0 )
	  {
	    // we have to look further in the compare cache for it
	    while ( true )
	    {
	      if ( ourItemset.isEqualTo ( theirItemset ) )
	      {
		break;
	      } // end if itemset is found
	      
	      // put the current their itemset in the their not found list
	      theirNotFound.add ( theirItemset );
	      
	      // see if the next one matches
	      theirItemset = dbCacheReader.getNextItemset();
	      
	    } // end while their itemset

	  } // end if we have to look more in the compare cache file
	    
	} // end try their itemset
	catch ( EOFException e )
	{
	  // we didn't find our itemset, add it to our not found list
	  notFound.add ( ourItemset );
	}
	
	// find the next one
	ourItemset = this.getNextItemset();
	
      } // end while our itemset
      
    } // end try our itemset
    catch ( EOFException e )
    {
      this.close();
      dbCacheReader.close();
    }

    return ( notFound );

  } // END public Vector compare ( DBCacheReader dbCacheReader )
  //    throws IOException, ClassNotFoundException

//********************************************************//

  /** 
   * For testing. Also contains a utility for comparing
   * two cache files. The itemsets from the first cache (-t)
   * file not found in the second (-c) are displayed.
   */

  public static void main ( String[] args )
  {
    if ( args.length < 2 )
    {
      System.out.println ( "DBCacheReader requires a file name of the cache " +
			   "to be read. (-t)" );
    } // end if no args

    String cacheString =  "";

    try
    {
      // need the -t file too
      cacheString =  Utils.getOption ( 't', args );
    }
    catch ( Exception e ) 
    {
      e.printStackTrace();
      System.out.println ( "\n" + e.getMessage() );
      
    } //end catch
    
    String compareCacheString = "";
    
    try
    {
      // should we compare itemsets?
      compareCacheString = Utils.getOption ( 'c', args );
    }
    catch ( Exception e ) 
    {
      e.printStackTrace();
      System.out.println ( "\n" + e.getMessage() );

    } //end catch

    if ( compareCacheString.length() != 0 )
    {
      System.out.println ( "Itemsets in " + cacheString + 
			   " not in " + compareCacheString + " : " );

      // try making db cache readers
      DBCacheReader cacheReader, compareCacheReader;

      try
      {
	cacheReader = new DBCacheReader ( cacheString );
	compareCacheReader = new DBCacheReader ( compareCacheString );

	System.out.println ( cacheReader.compare ( compareCacheReader ) );

      } // end try making db cache readers for compare purposes
      catch ( Exception e )
      {
	System.out.println ( e );
      }

      // all done
      return;

    } // end if we should compare files

    try
    {
      DBCacheReader dbcache = new DBCacheReader ( cacheString );
      
      try
      {
	while ( true )
	{
	  System.out.println ( dbcache.getNextItemset() );
	}
      }
      catch ( EOFException e )
      {
	dbcache.close();
      }
    }
    catch ( Exception e )
    {
      System.out.println ( e );
    }

  } // END public static void main ( String[] args )

//********************************************************//

} // END public class DBCacheReader
