/*
This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or (at
your option) any later version.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
USA
*/

/*
  Maintainance log started on January 3rd, 2004

  January 3, 2004   - created Logger class -kap

  February 22, 2004 - added metrics min-event-length, max-event-length,
                      min-per-event-length, and max-per-event-length -kap
*/

package wpi.associations.arminerSequence;

import wpi.timesequence.TSUtility;

import java.io.*;
import java.util.Arrays;
import java.util.Vector;

/**
 * Handles logging metrics about the Apriori Sets And
 * Sequences algorithm during run time. These metrics are
 * written to a file in ARFF format.
 *
 * @author Keith A. Pray (kap@wpi.edu)
 */

public class Logger
{

//********************************************************//
// ----- Data Members -----
//********************************************************//  

  /** The name of the log file to write to. */
  private String logStatsFileName;

  public static String defaultLogStatsFileName = "asas-log.arff";

  /* To add a metric see the getMetricInfo method and list
   of static final metric codes */

  /** An array of the metrics and their data types captured in 
      logging mode */
  private String[] logMetricNames;

  /** A list of the values to be stored in the log file */
  private StringBuffer[] logMetricValues;

  /** Specifies the name of a metric when retrieving information */
  public static final int NAME = 0;

  /** Specifies the data type of a metric when retreiving information */
  public static final int TYPE = 1;

  // A list of the metrics this logger can handle follows.
  // If you wish to add a metric, add it to the end of this
  // list giving it the next number. Add the string name and
  // data type of the metric to the getLogMetricInfo method.

  /** Number of seconds to generate candidates item sets of size k */
  public static final int SECONDS_TO_GENERATE_CANDIDATES     = 0;

  /** Number of seconds to count support of candidate item sets of size k */
  public static final int SECONDS_TO_COUNT_SUPPORT           = 1;

  /** Number of seconds for the entire frequent item set mining */
  public static final int TOTAL_SECONDS_TO_MINE              = 2;

  /** Number of candidate item sets of size k generated */
  public static final int NUM_CANDIDATES_GENERATED           = 3;

  /** Number of unique item sets candidates of size k generated */
  public static final int NUM_UNIQUE_CANDIDATES_GENERATED    = 4;

  /** Number of frequent item sets of size k found */
  public static final int NUM_FREQUENT_ITEMSETS_FOUND        = 5;

  /** Number of frequent unique item sets of size k found */
  public static final int NUM_UNIQUE_FREQUENT_ITEMSETS_FOUND = 6;

  /** Number of times a candidate item sets did not have to be compared 
      to instances in the data set because the candidate was removed from
      the candidate list when it was clear it could not be considered
      frequent due to the number of instances the candidate was already
      compared to */
  public static final int NUM_COMPARISONS_SAVED_REMOVING_CANDIDATES_DURING_SUPPORT_COUNTING = 7;

  /** Number of event items of the same type allowed to appear in a rule */
  public static final int NUM_EVENTS_OF_SAME_TYPE_ALLOWED    = 8;

  /** Number of attributes present in the data set */
  public static final int NUM_ATTRIBUTES                     = 9;

  /** Number of event attributes present in the data set */
  public static final int NUM_EVENT_ATTRIBUTES               = 10;

  /** Number of times events occur in the data set */
  public static final int NUM_EVENT_OCCURRENCES              = 11;

  /** Average length of a time line in the data set */
  public static final int AVERAGE_LENGTH_OF_TIMELINE         = 12;

  /** Number of instances in the data set */
  public static final int NUM_INSTANCES                      = 13;

  /** Specifies if item set prefix tree used instead of vector.
      "true" if the tree is used, "false" otherwise */
  public static final int USE_ITEMSET_PREFIX_TREE            = 14;

  /** Specifies if previous duplicate hash used.
      "true" if the tree is used, "false" otherwise */
  public static final int USE_PREVIOUS_DUPLICATE_HASH        = 15;

  /** Number of seconds to generate association rules
      from frequent item sets */
  public static final int SECONDS_TO_GENERATE_RULES          = 16;

  /** Number of seconds to calculate find confidence of
      association rules. */
  public static final int SECONDS_TO_FIND_CONFIDENCE         = 17;

  /** Number of association rules found. */
  public static final int NUM_RULES                          = 18;

  /** Relation name of the data set. */
  public static final int RELATION_NAME                      = 19;

  /** Date and time when the mining started. */
  public static final int TIMESTAMP                          = 20;

  /** Support metric. */
  public static final int SUPPORT                            = 21;

  /** Confidence metric. */
  public static final int CONFIDENCE                         = 22;

  /** Total mining seconds per frequent item set found. */
  public static final int SECONDS_PER_FREQUENT_ITEMSET       = 23;

  /** Total number of frequent item set found */
  public static final int TOTAL_NUM_FREQUENT_ITEMSETS        = 24;

  /** Average number of event occurrences per event attribute */
  public static final int NUM_OCCURRENCES_PER_ATTRIBUTE      = 25;

  /** Minimum length of the events. */
  public static final int MIN_EVENT_LENGTH                   = 26;

  /** Maximum length of the events. */
  public static final int MAX_EVENT_LENGTH                   = 27;

  /** Minimum length per event. */
  public static final int MIN_PER_EVENT_LENGTH                   = 28;

  /** Maximum length per event. */
  public static final int MAX_PER_EVENT_LENGTH                   = 29;

  /** Specifies debug info level 
   * 0: no debug info
   * 1: input to methods
   * 2: and output from methods
   * 3: and all sorts of stuff 
   */
  private static final int debug = 0;

//********************************************************//
// ----- Methods-----
//********************************************************//

/** 
 * Constructor that allows the log file name to be 
 * specified. 
 */

  public Logger ( String fileName )
  {
    // first initialize the metric names array
    initLogMetricNames();

    switch ( debug ) {
    case 3: // case 0:
      System.out.println ( "Logger.constructor: " +
			   "logMetricNames: " + 
			   Arrays.asList ( logMetricNames ) );
    } // end switch debug

    // now we can set the file name
    setLogStatsFileName ( fileName );

    // initialize the metric values array
    initLogMetricValues();

  } // END public Logger ( String fileName )
  
//********************************************************//

/** 
 * Constructor that uses the default log file name.
 */

  public Logger()
  {
    setLogStatsFileName ( defaultLogStatsFileName );

  } // END public Logger()
  
//********************************************************//

  /**
   * Sets the name of the file in which the statistics
   * will be logged to. If the file doesn't exist one
   * will be created.
   *
   * @param fileName
   * name of the file to store statistics
   */

  public void setLogStatsFileName ( String fileName )
  {
    logStatsFileName = fileName;

    // need a file reader
    FileReader fileReader = null;
    
    try
    {
      fileReader = new FileReader ( logStatsFileName );
    }
    catch ( FileNotFoundException ex )
    {
      // no existing file, must make a new one
      FileWriter fileWriter = null;

      // try to open the file for writing
      try
      {
	// open the file writer
	fileWriter = new FileWriter ( logStatsFileName );
	
	// let's use a buffered writer to make it easy
	BufferedWriter bufferedWriter = new BufferedWriter ( fileWriter );
	
	// we need a relation
	bufferedWriter.write ( "@relation asas-log" );
	bufferedWriter.newLine();
	bufferedWriter.newLine();
	
	// we need to list the attributes
	
	for ( int i = 0; i < logMetricNames.length; i++ )
	{
	  bufferedWriter.write ( "@attribute " + logMetricNames [ i ] +
				 " " + getLogMetricType ( i ) );
	  bufferedWriter.newLine();
	  
	} // end for each log metric
	
	// define start of data
	bufferedWriter.newLine();
	bufferedWriter.write ( "@data" );
	bufferedWriter.newLine();
	bufferedWriter.newLine();
	
	// make sure the buffer gets written to the file
	bufferedWriter.flush();
	bufferedWriter.close();
	
      } // end try
      catch ( IOException ex2 )
      {
	
      } // end catch
      
    } // end catch no log file present, gotta make one
    
  } // END public void setLogStatsFileName ( String fileName )
  
//********************************************************//

  /**
   * Get the list of metrics that can be stored in the log
   * file.
   *
   * @return an array of the names of the metrics
   */

  public String[] getLogMetricNames()
  {
    return ( (String[])logMetricNames.clone() );

  } // END public getLogMetricNames()
  
//********************************************************//

  /**
   * Get information about the log metric with the specified
   * integer code. An empty string is returned if there is
   * no metric associated with the code. The information
   * returned is specified by using a code.
   *
   * @param code
   * the integer code representing a mteric
   *
   * @param info
   * Logger.NAME for the metric name<br>
   * Logger.TYPE for the metric data type
   *
   * @return the metric name, empty string if code not 
   * found
   */

  public String getLogMetricInfo ( int code, int info )
  {
    switch ( code )
    {
    case SECONDS_TO_GENERATE_CANDIDATES:     // 0
      switch ( info ) {
      case NAME: return ( "seconds-to-generate-candidates" );
      case TYPE: return ( "string" );
      } // end switch info
    case SECONDS_TO_COUNT_SUPPORT:           // 1
      switch ( info ) {
      case NAME: return ( "seconds-to-count-support" );
      case TYPE: return ( "string" );
      } // end switch info
      case TOTAL_SECONDS_TO_MINE:            // 2
      switch ( info ) {
      case NAME: return ( "total-seconds-to-mine" );
      case TYPE: return ( "real" );
      } // end switch info
    case NUM_CANDIDATES_GENERATED:           // 3
      switch ( info ) {
      case NAME: return ( "#-candidates-generated" );
      case TYPE: return ( "string" );
      } // end switch info
    case NUM_UNIQUE_CANDIDATES_GENERATED:    // 4
      switch ( info ) {
      case NAME: return ( "#-unique-candidates-generated" );
      case TYPE: return ( "string" );
      } // end switch info
    case NUM_FREQUENT_ITEMSETS_FOUND:        // 5
      switch ( info ) {
      case NAME: return ( "#-frequent-itemsets-found" );
      case TYPE: return ( "string" );
      } // end switch info
    case NUM_UNIQUE_FREQUENT_ITEMSETS_FOUND: // 6
      switch ( info ) {
      case NAME: return ( "#-unique-frequent-itemsets-found" );
      case TYPE: return ( "string" );
      } // end switch info
    case NUM_COMPARISONS_SAVED_REMOVING_CANDIDATES_DURING_SUPPORT_COUNTING: // 7
      switch ( info ) {
      case NAME: return ( "#-comparisons-saved-removing-candidates-during-support-counting" );
      case TYPE: return ( "string" );
      } // end switch info
    case NUM_EVENTS_OF_SAME_TYPE_ALLOWED:    // 8
      switch ( info ) {
      case NAME: return ( "#-events-of-same-type-allowed" );
      case TYPE: return ( "real" );
      } // end switch info
    case NUM_ATTRIBUTES:                     // 9
      switch ( info ) {
      case NAME: return ( "#-attributes" );
      case TYPE: return ( "real" );
      } // end switch info
    case NUM_EVENT_ATTRIBUTES:               // 10
      switch ( info ) {
      case NAME: return ( "#-event-attributes" );
      case TYPE: return ( "real" );
      } // end switch info
    case NUM_EVENT_OCCURRENCES:              // 11
      switch ( info ) {
      case NAME: return ( "#-event-occurrences" );
      case TYPE: return ( "real" );
      } // end switch info
    case AVERAGE_LENGTH_OF_TIMELINE:         // 12
      switch ( info ) {
      case NAME: return ( "average-length-of-timeline" );
      case TYPE: return ( "real" );
      } // end switch info
    case NUM_INSTANCES:                      // 13
      switch ( info ) {
      case NAME: return ( "#-instances" );
      case TYPE: return ( "real" );
      } // end switch info
    case USE_ITEMSET_PREFIX_TREE:            // 14
      switch ( info ) {
      case NAME: return ( "use-itemset-prefix-tree" );
      case TYPE: return ( "string" );
      } // end switch info
    case USE_PREVIOUS_DUPLICATE_HASH:            // 15
      switch ( info ) {
      case NAME: return ( "use-previous-duplicate-hash" );
      case TYPE: return ( "string" );
      } // end switch info
    case SECONDS_TO_GENERATE_RULES:            // 16
      switch ( info ) {
      case NAME: return ( "seconds-to-generate-rules" );
      case TYPE: return ( "string" );
      } // end switch info
    case SECONDS_TO_FIND_CONFIDENCE:            // 17
      switch ( info ) {
      case NAME: return ( "seconds-to-find-confidence" );
      case TYPE: return ( "string" );
      } // end switch info
    case NUM_RULES:            // 18
      switch ( info ) {
      case NAME: return ( "num-rules" );
      case TYPE: return ( "string" );
      } // end switch info
    case RELATION_NAME:            // 19
      switch ( info ) {
      case NAME: return ( "relation-name" );
      case TYPE: return ( "string" );
      } // end switch info
    case TIMESTAMP:            // 20
      switch ( info ) {
      case NAME: return ( "timestamp" );
      case TYPE: return ( "string" );
      } // end switch info
    case SUPPORT:            // 21
      switch ( info ) {
      case NAME: return ( "support" );
      case TYPE: return ( "string" );
      } // end switch info
    case CONFIDENCE:            // 22
      switch ( info ) {
      case NAME: return ( "confidence" );
      case TYPE: return ( "real" );
      } // end switch info
      case SECONDS_PER_FREQUENT_ITEMSET:      //23;
      switch ( info ) {
      case NAME: return ( "seconds-per-frequent-itemset" );
      case TYPE: return ( "real" );
      } // end switch info
      case TOTAL_NUM_FREQUENT_ITEMSETS:      //24;
      switch ( info ) {
      case NAME: return ( "total-num-frequent-itemsets" );
      case TYPE: return ( "real" );
      } // end switch info
      case NUM_OCCURRENCES_PER_ATTRIBUTE:      //25;
      switch ( info ) {
      case NAME: return ( "num-occurrences-per-attribute" );
      case TYPE: return ( "real" );
      } // end switch info
      case MIN_EVENT_LENGTH:      //26;
      switch ( info ) {
      case NAME: return ( "min-event-length" );
      case TYPE: return ( "real" );
      } // end switch info
      case MAX_EVENT_LENGTH:      //27;
      switch ( info ) {
      case NAME: return ( "max-event-length" );
      case TYPE: return ( "real" );
      } // end switch info
      case MIN_PER_EVENT_LENGTH:      //28;
      switch ( info ) {
      case NAME: return ( "min-per-event-length" );
      case TYPE: return ( "string" );
      } // end switch info
      case MAX_PER_EVENT_LENGTH:      //29;
      switch ( info ) {
      case NAME: return ( "max-per-event-length" );
      case TYPE: return ( "string" );
      } // end switch info
    default:
      return ( "" );

    } // end switch code

  } // END public String getLogMetricInfo ( int code, 
  //  int info )
  
//********************************************************//

  /**
   * Get the name of the log metric with the specified
   * integer code. An empty string is returned if there is
   * no metric associated with the code.
   *
   * @param code
   * the integer code representing a mteric
   *
   * @return the metric name, empty string if code not 
   * found
   */

  public String getLogMetricName ( int code )
  {
    return ( getLogMetricInfo ( code, NAME ) );

  } // END public String getLogMetricName ( int code )
  
//********************************************************//

  /**
   * Get the data type of the log metric with the specified
   * integer code. An empty string is returned if there is
   * no metric associated with the code.
   *
   * @param code
   * the integer code representing a mteric
   *
   * @return the metric data type, empty string if code not 
   * found
   */

  public String getLogMetricType ( int code )
  {
    return ( getLogMetricInfo ( code, TYPE ) );

  } // END public String getLogMetricType ( int code )
  
//********************************************************//

  /**
   * Initializes the array of metric names.
   */

  private void initLogMetricNames()
  {
    int code = 0;
    Vector tempNames = new Vector();
    String tempString = "";

    while ( true )
    {
      tempString = getLogMetricName ( code );

      if ( tempString.equals ( "" ) )
      {
	break;

      } // end if no metric for code
      else 
      {
	tempNames.add ( tempString );

      } // end else there was a metric

      code++;

    } // end while true

    logMetricNames = new String [ tempNames.size() ];

    for ( code = 0; code < logMetricNames.length; code++ )
    {
      logMetricNames [ code ] = (String)tempNames.elementAt ( code );

    } // end for each log metric name

  } // END private void initLogMetricNames()
  
//********************************************************//

  /**
   * Initializes the array of log metric values.
   */

  private void initLogMetricValues()
  {
    logMetricValues = new StringBuffer [ logMetricNames.length ];

    // have to use a loop to initialize the array since
    // a new StringBuffer object must be assigned to each
    // index
    for ( int i = 0; i < logMetricValues.length; i++ )
    {
      logMetricValues [ i ] = new StringBuffer ( "" );
      
    } // end for each value
    
  } // END private void initLogMetricValues()
  
//********************************************************//

  /**
   * Returns the value saved for the specified metric.
   *
   * @param code
   * code for the metric
   *
   * @return the string stored for the metric, 
   * null if the code is not valid
   */

  public String getValue ( int code )
  {
    // is the code valid
    if ( ( code < logMetricValues.length ) && ( code > -1 ) )
    {
      return ( logMetricValues [ code ].toString() );

    } // end if code valid

    // not a valid code I guess
    return ( null );

  } // END public String getValue ( int code )

//********************************************************//

  /**
   * Add a metric value to the log. If a value already
   * exists for a metric it will be assumed that the
   * metric can be recorded as a time sequence. The new
   * value will be seperated from the previous value by
   * the time sequence delimiter specified in 
   * wpi.timesequence.TSUtility.
   *
   * @param code
   * code for the metric
   *
   * @param value
   * value to log for the metric
   */

  public void logValue ( int code, String value )
  {
    // is the code valid
    if ( code < logMetricValues.length )
    {
      // is there already a value?
      if ( logMetricValues [ code ].toString().equals ( "" ) )
      {
	switch ( debug ) {
	case 3: case 2: case 1: //case 0:
	  System.out.println ( "Logger.logValue: " +
			       "logging " + 
			       getLogMetricName ( code ) + "(" + code + ")" +
			       " = " + value );
	} // end switch debug

	(logMetricValues [ code ]).append ( value );

      } // end if no value yet
      else
      {
	(logMetricValues [ code ]).append ( TSUtility.TIME_SEQUENCE_DELIMITER +
					    value );
      } // end else value exists

    } // end if code valid

  } // END public void logValue ( int code, String value )
  
//********************************************************//

  /** 
   * Saves the logged values to the log file.
   */

  public void saveLog()
  {
    switch ( debug ) {
    case 3: case 2: case 1: //case 0:
      System.out.println ( "Logger.saveLog: " +
			   "logMetricValues: " + 
			   Arrays.asList ( logMetricValues ) );
    } // end switch debug

    // construct the string to write to the log file
    StringBuffer saveString = new StringBuffer ( "" );

    // are there any values to log?
    switch ( logMetricValues.length )
    {
    case 0:
      // nothing to do
      return;

    default:
      // only have to write this one value
      saveString.append ( "'" + logMetricValues [ 0 ] + "'" );
      break;

    } // end switch logMetricValues

    // starting at the second value, if there is one
    // append the values seperated by commas
    for ( int i = 1; i < logMetricValues.length; i++ )
    {
      saveString.append ( ",'" + logMetricValues [ i ] + "'" );

      switch ( debug ) {
      case 3: case 2: case 1: //case 0:
	System.out.println ( "Logger.saveLog: " +
			     "added value: " + ",'" +
			     logMetricValues [ i ] + "'" );
			     
      } // end switch debug

    } // end for each log metric value

    // ok, write the line to the file
    FileWriter fileWriter = null;

    // try to open the file for writing
    try
    {
      // open the file writer, be sure to specify true for the
      // optional append
      fileWriter = new FileWriter ( logStatsFileName, true );
      
      // let's use a buffered writer to make it easy
      BufferedWriter bufferedWriter = new BufferedWriter ( fileWriter );
      
      // we need a relation
      bufferedWriter.write ( saveString.toString() );
      bufferedWriter.newLine();    
	
      // make sure the buffer gets written to the file
      bufferedWriter.flush();
      bufferedWriter.close();

      switch ( debug ) {
      case 3: case 2: case 1: //case 0:
	System.out.println ( "Logger.saveLog: " +
			     "saveString: " + saveString );
			     
      } // end switch debug
	
      } // end try
      catch ( IOException ex )
      {
	System.err.println ( "Logger.saveLog: " +
			     "problem saving values: " +
			     Arrays.asList ( logMetricValues ) +
			     " to log file." );
	System.err.println ( ex.getMessage() );
	
      } // end catch

  } // END public void saveLog()

//********************************************************//

  /**
   * Clears the metrics logged thus far.
   */

  public void clear()
  {
    // set each value to empty string
    for ( int i = 0; i < logMetricValues.length; i++ )
    {
      logMetricValues [ i ].setLength ( 0 );
      
    } // end for each value

  } // END public void clear()
  
//********************************************************//

  /**
   * Check if the Logger is clear. 
   *
   * @returns true if no values currently logged,
   * false otherwise
   */

  public boolean isClear()
  {
    // check each value for empty string
    for ( int i = 0; i < logMetricValues.length; i++ )
    {
      if ( logMetricValues [ i ].length() > 0 )
      {
	// then there must be a value logged
	return ( false );

      } // end if length more than zero
      
    } // end for each value

    // I guess we found no values so we must be clear
    return ( true );

  } // END public boolean isClear()
  
//********************************************************//

} // END public class Logger
