package wpi.filters;

import weka.core.*;
import weka.filters.*;

import java.util.*;
import java.io.*;

/**
 * <p>
 * This filter converts a string attribute to a set of
 * yes & no nominal attributes.  The new attributes will be created
 * for every unique entry separated by delimiter in the string.
 * Each of the new attributes will be named as 'attributeName-entryValue'.<br>
 * (Note: The string attribute with no delimiter in any of its instance is
 * not affected.)
 * </p>
 *
 * <p>Valid options are:</p>
 *
 * <p>
 * -D delimiters
 * Specify the delimiters of the set.
 * (default "|")<br>
 *
 * -N nominal_values
 * Specify the nominal values used to classify an attribute.
 * Valid values are "yesno", "truefalse", and "onezero".
 * (default "yesno")<br>
 *
 * -R range
 * Specify the range of filtering.
 * e.g. 'first-last', or '1-2,6,8-last'.
 * (default "all")<br>
 *
 * -I
 * Use this flag to inverse the range.<br>
 * </p>
 *
 * @author Takeshi Kawato (takeshi@wpi.edu)
 * @version $Revision: 0.1 $
 **/
public class AttributeSetFilter extends Filter implements OptionHandler
{
  /**
   * The static integer used to define nominal values of yes and no.
   **/
  public static int YESNO = 0;

  /**
   * The static integer used to define nominal values of true and false.
   **/
  public static int TRUEFALSE = 1;

  /**
   * The static integer used to define nominal values of 1 and 0.
   **/
  public static int ONEZERO = 2;

  /**
   * The integer defining current nominal values.
   **/
  protected int m_nom_values = 0;
  
  /**
   * The string definining a set.
   **/
  protected String SET = "|SET";

  /**
   * The string defining a non-set.
   **/
  protected String NONSET = "|NONSET";

  /**
   * Stores the input format.
   **/
  protected Instances m_InputFormat = null;

  /**
   * Stores the output format.
   **/
  protected Instances m_OutputFormat = null;

  /**
   * Stores the set format.<br>
   * e.g. [SET][entry][entry][entry][NONSET][SET][entry]
   **/
  protected FastVector m_SetFormat = null;

  /**
   * Stores the nominal value for an instance with an entry.
   **/
  protected String m_nom_yes = "yes";

  /**
   * Stores the nominal value for an instance without an entry.
   **/
  protected String m_nom_no = "no";

  /**
   * Stores the delimeter(s) used for separating entries in a set.
   **/
  protected String m_delimiters = "|";

  /**
   * Stores the range of filtering as a string.
   **/
  protected String m_range = "first-last";

  /**
   * Stores whether the range of filtering should be inversed, or not.
   **/
  protected boolean m_rangeInverse = false;

  /**
   * Stores the initial hash capacity for a hash used internally.
   **/
  protected int m_hashCapacity = 10000;

  /**
   * Define possible nominal values.
   **/
  public static final Tag [] TAGS_NOMINAL_VALUES = {
    new Tag(YESNO, "yes or no"),
    new Tag(TRUEFALSE, "true or false"),
    new Tag(ONEZERO, "1 or 0")};


  /**
   * Returns a string describing this filter
   *
   * @return a description of the filter suitable for
   * displaying in the explorer/experimenter gui.
   **/
  public String globalInfo()
  {
    return "This filter converts a string attribute to a set of " +
           "yes & no nominal attributes.  The new attributes will be created " +
           "for every unique entry separated by delimiter in the string.  " +
           "Each of the new attributes will be named as " +
           "'attributeName-entryValue'.\n" +
           "(Note: The string attribute with no delimiter in any of its " +
           "instance is not affected.)";
  }

  /**
   * Sets the format of the input instances.
   *
   * @param instanceInfo an Instances object containing the input instance
   * structure (any instances contained in the object are ignored - only the
   * structure is required).
   * @return true if the outputFormat may be collected immediately.
   * @exception Exception if the inputFormat can't be set successfully.
   **/
  public boolean setInputFormat(Instances instanceInfo) throws Exception
  {
    super.setInputFormat(instanceInfo);

    m_InputFormat = new Instances(instanceInfo, 0);

    m_NewBatch = true;
    m_OutputFormat = null;
    m_SetFormat = null;

    //conversion can take place only after batch has finished.
    return false;
  }

  /**
   * Input an instance for filtering. This filter requires all
   * instances be read before producing output.
   *
   * @param instance the input instance
   * @return true if the filtered instance may now be collected with output().
   * @exception IllegalStateException if no input format has been defined.
   * @exception Exception if the input instance was not of the correct format 
   * or if there was a problem with the filtering.
   **/
  public boolean input(Instance instance) throws Exception
  {
    if(m_InputFormat == null)
    {
      //input format is not defined.
      throw new IllegalStateException("No input instance format defined");
    }

    if(m_NewBatch)
    {
      //this is a new batch, reset the queue.
      resetQueue();
      m_NewBatch = false;
    }

    if(m_OutputFormat == null)
    {
      //output format is not defined.

      //add the instance for output format creation.
      m_InputFormat.add(instance);

      return false;
    }
    else
    {
      //output format is defined.

      //convert the instance to the output format.
      convertInstance(instance);

      return true;
    }
  }

  /**
   * Signify that this batch of input to the filter is finished. If
   * output() may now be called to retrieve the filtered instances.
   *
   * @return true if there are instances pending output.
   * @exception NullPointerException if no input structure has been defined,
   * @exception Exception if there was a problem finishing the batch.
   **/
  public boolean batchFinished() throws Exception
  {
    if(m_InputFormat == null)
    {
      //input format is not defined.
      throw new Exception("No input instance format defined");
    }

    if(m_OutputFormat == null)
    {
      //output format is not defined.

      //create output format from given input format.
      convertInstanceFormat(m_InputFormat);

      for(int i = 0; i < m_InputFormat.numInstances(); i++)
      {
        //for each instance in input format:

        //convert the instance to output the format.
        convertInstance(m_InputFormat.instance(i));
      }
    }

    m_NewBatch = true;

    return (numPendingOutput() != 0);
  }

  /**
   * Returns the tip text for this property.
   *
   * @return tip text for this property suitable for
   * displaying in the explorer/experimenter gui.
   **/
  public String delimitersTipText()
  {
    return "Set the delimiters of the set.\n" +
           "Each character in the field will be treated as a delimeter.";
  }

  /**
   * Set the delimiters of the set.
   * Each character in the field will be treated as a delimeter.
   *
   * @param delimeter a string representing the delimeter(s).
   * @exception Exception if an invalid delimeter is supplied.
   **/
  public void setDelimiters(String delimiters) throws Exception
  {
    if(delimiters == null || delimiters.equals(""))
    {
      //delimeter is not valid.
      throw new Exception("Delimeter must be specified.");
    }

    //set the delimiters.
    m_delimiters = delimiters;

    //set the string defining a set.
    SET = m_delimiters + "SET";

    //set the string defining a non-set.
    NONSET = m_delimiters + "NONSET";
  }

  /**
   * Get the current delimeter(s).
   *
   * @return a string containing the current delimeter(s).
   **/
  public String getDelimiters()
  {
    return m_delimiters;
  }

  /**
   * Returns the tip text for this property.
   *
   * @return tip text for this property suitable for
   * displaying in the explorer/experimenter gui.
   **/
  public String nominalValuesTipText()
  {
    return "Set the nominal values used classify an attribute.";
  }

  /**
   * Set the nominal values used to classify an attribute.
   *
   * @param nominal_values the nominal values used to classify an attribute.
   **/
  public void setNominalValues(SelectedTag nominal_values)
  {
    if(nominal_values.getTags() == TAGS_NOMINAL_VALUES)
    {
      if(nominal_values.getSelectedTag().getID() == YESNO)
      {
        //set nominal values to yes or no
        m_nom_values = YESNO;
        m_nom_yes = "yes";
        m_nom_no = "no";
      }
      else if(nominal_values.getSelectedTag().getID() == TRUEFALSE)
      {
        //set nominal values to true or false
        m_nom_values = TRUEFALSE;
        m_nom_yes = "true";
        m_nom_no = "false";
      }
      else if(nominal_values.getSelectedTag().getID() == ONEZERO)
      {
        //set nominal values to 1 or 0
        m_nom_values = ONEZERO;
        m_nom_yes = "1";
        m_nom_no = "0";
      }
    }
  }

  /**
   * Get the current nominal values used to classify an attribute.
   *
   * @return a selected nominal values.
   **/
  public SelectedTag getNominalValues()
  {
    return new SelectedTag(m_nom_values, TAGS_NOMINAL_VALUES);
  }

  /**
   * Returns the tip text for this property.
   *
   * @return tip text for this property suitable for
   * displaying in the explorer/experimenter gui.
   **/
  public String rangeTipText()
  {
    return "Set the range of filtering.\n" +
           "e.g. 'first-last', or '1-2,6,8-last'.\n" +
           "Note: This filter will ignore any non-set attribute " +
           "(The numeric, nominal, or string attribute with no delimeter " +
           "in any of its instances).  One does not have to specify specific " +
           "range except for extraneous cases (Cases where some instances of " +
           "a non-set string attribute containing a delimeter).";
  }

  /**
   * Set the range of filtering.<br>
   * Note: This filter will ignore any non-set attribute
   * (The numeric, nominal, or string attribute with no delimeter
   * in any of its instances).  One does not have to specify specific
   * range except for extraneous cases (Cases where some instances of
   * a non-set string attribute containing a delimeter).
   *
   * @param range a string representing the range.  Since the string will
   * typically come from a user, attributes are indexed from 1.<br>
   * e.g. 'first-last', or '1-2,6,8-last'.
   * @exception Exception if an invalid range is supplied
   **/
  public void setRange(String range) throws Exception
  {
    if(range == null || range.equals(""))
    {
      //range is not valid.

      throw new Exception("Range must be specified.");
    }

    try
    {
      Range r = new Range(range);
      m_range = range;
    }
    catch(Exception e)
    {
      throw e;
    }
  }

  /**
   * Get the current range of filtering.
   *
   * @return a string representing the current range.
   **/
  public String getRange()
  {
    return m_range;
  }

  /**
   * Returns the tip text for this property.
   *
   * @return tip text for this property suitable for
   * displaying in the explorer/experimenter gui.
   **/
  public String rangeInverseTipText()
  {
    return "Use this flag to inverse the range.";
  }

  /**
   * Inverses the range of filtering.
   *
   * @param rangeInverse true, if the range is to be inversed. false, otherwise.
   **/
  public void setRangeInverse(boolean rangeInverse)
  {
    m_rangeInverse = rangeInverse;
  }

  /**
   * Get whether the current range of filtering is to be inversed or not.
   *
   * @return true, if the range is to be inversed. false, otherwise.
   **/
  public boolean getRangeInverse()
  {
    return m_rangeInverse;
  }

  /**
   * Convert an instance format to output instance format.
   *
   * @param instanceFormat the instance format to convert.
   **/
  protected void convertInstanceFormat(Instances instanceFormat)
  {
    //create a vector of nominal values.
    FastVector nom_values = new FastVector();
    nom_values.addElement(m_nom_yes);
    nom_values.addElement(m_nom_no);

    //create a range to work on.
    Range range = new Range(m_range);
    range.setUpper(instanceFormat.numAttributes());
    range.setInvert(m_rangeInverse);

    //create new output format.
    m_OutputFormat = new Instances(instanceFormat.relationName(), new FastVector(), 0);

    //create new set format.
    m_SetFormat = new FastVector();

    for(int i = 0, j = 0; i < instanceFormat.numAttributes(); i++)
    {
      //for each attribute in given instance format:

      if(isAttributeSet(instanceFormat.attribute(i)) && range.isInRange(i))
      {
        //attribute is a set.

        //update the set format.
        m_SetFormat.addElement(SET);

        //create a hash that holds each unique entry in the set.
        Hashtable hash = createHash(instanceFormat.attribute(i));

        Enumeration enum = hash.keys();

        while(enum.hasMoreElements())
        {
          //for each unique entry in the set:
          String attributeName = (String)enum.nextElement();

          //update the set format.
          m_SetFormat.addElement(hash.get(attributeName));

          //create a new attribute.
          m_OutputFormat.insertAttributeAt(new Attribute(attributeName, nom_values), j);
          j++;
        }
      }
      else
      {
        //attribute is not a set.

        //update the set format.
        m_SetFormat.addElement(NONSET);

        //create a copy of the attribute.
        m_OutputFormat.insertAttributeAt((Attribute)instanceFormat.attribute(i).copy(), j);
        j++;
      }
    }

    //set output format.
    setOutputFormat(m_OutputFormat);
  }

  /**
   * Convert an instance to the output format.
   * The converted instance is added to the end of the output queue.
   *
   * @param instance the instance to convert.
   **/
  protected void convertInstance(Instance instance)
  {
    //create a new instance.
    Instance newInstance = new Instance(m_OutputFormat.numAttributes());
    newInstance.setDataset(new Instances(m_OutputFormat));

    for(int i = 0, j = 0, k = 0; i < m_InputFormat.numAttributes(); i++)
    {
      //for each attribute in input format:

      if(m_SetFormat.elementAt(k).equals(NONSET))
      {
        //the attribute is not a set.

        //create a copy of the attribute value.
        if(instance.attribute(i).isNumeric())
        {
          newInstance.setValue(j, instance.value(i));
          j++;
        }
        else
        {
          newInstance.setValue(j, instance.stringValue(i));
          j++;
        }

        k++;
      }
      else if(m_SetFormat.elementAt(k).equals(SET))
      {
        //the attribute is a set.
        k++;

        while(m_SetFormat.elementAt(k) != null &&
              !m_SetFormat.elementAt(k).equals(SET) &&
              !m_SetFormat.elementAt(k).equals(NONSET))
        {
          //until we encounter a new SET or NONSET entry:
          String entry = (String)m_SetFormat.elementAt(k);
          String value = m_nom_no;

          if(instance.isMissing(i))
          {
            //attribute is missing, treat it as empty set.

            //entry is not in the attribute set.
            value = m_nom_no;
          }
          else if(entryIsInSet(entry, instance.stringValue(i)))
          {
            //entry is in the attribute set.
            value = m_nom_yes;
          }
          else
          {
            //entry is not in the attribute set.
            value = m_nom_no;
          }

          //create a new attribute.
          newInstance.setValue(j, value);
          j++;

          k++;
        }
      }
    }

    //push new instance into the queue.
    push(newInstance);
  }

  /**
   * Tests whether the attribute is a set.
   *
   * @param attribute the attribute to test.
   * @return true, if the attribute is a set. false, otherwise.
   **/
  protected boolean isAttributeSet(Attribute attribute)
  {
    if(attribute.isString())
    {
      //attribute is a string, which could be a set.

      for(int i = 0; i < attribute.numValues(); i++)
      {
        //for each attribute value:

        //create a string tokenizer with given delimiter.
        StringTokenizer st = new StringTokenizer(attribute.value(i), m_delimiters);

        if(st.countTokens() > 1)
        {
          //there was at least one string containing a set.
          return true;
        }
      }
    }

    //attribute is not a set.
    return false;
  }

  /**
   * Creates a hash of unique entries in a set attribute.
   *
   * @param attribute the attribute to create hash from.
   * @return a Hashtable holding unique entries (key as an attribute name for
   * the unique entry and value as an entry value itself).
   **/
  protected Hashtable createHash(Attribute attribute)
  {
    //create a hash for holding unique entry in attribute value.
    Hashtable hash = new Hashtable(m_hashCapacity);

    for(int i = 0; i < attribute.numValues(); i++)
    {
      //for each attribute value:

      //create a string tokenizer with given delimiter.
      StringTokenizer st = new StringTokenizer(attribute.value(i), m_delimiters);

      while(st.hasMoreTokens())
      {
        //for each entry in a attribute value:
        String entry = (String)st.nextToken();

        //create an attribute name for the entry.
        String attributeName = attribute.name() + "-" + entry;

        //put the attribute name and entry set into the hash.
        hash.put(attributeName, entry);
      }
    }

    return hash;
  }

  /**
   * Tests whether the entry is in the given set.
   *
   * @param entry the entry to test.
   * @param set the set to test.
   * @return true, if the entry is in the set. false, otherwise.
   **/
  protected boolean entryIsInSet(String entry, String set)
  {
    //create a string tokenizer with given delimiter.
    StringTokenizer st = new StringTokenizer(set, m_delimiters);

    while(st.hasMoreTokens())
    {
      //for each entry in the set:
      String set_entry = st.nextToken();

      //test whether the entry equals to the given entry.
      if(set_entry.equals(entry))
      {
        //the entry was in the set.
        return true;
      }
    }

    return false;
  }

  /**
   * Returns the string inside matching brackets.
   * This method assumes that the string starts with a starting bracket.
   *
   * @param startBracket the starting bracket
   * @param endBracket the ending bracket
   * @param text the text to examine
   *
   * @return the string inside matching brackets.
   **/
  public String getInsideBracket(char startBracket, char endBracket, String text)
  {
    int count = 0;
    int sbcount = 0;
    int ebcount = 0;

    for(count = 0; count < text.length(); count++)
    {
      if(text.charAt(count) == startBracket)
      {
        //the character is starting bracket.

        //increment the number of starting bracket encountered.
        sbcount++;
      }

      if(text.charAt(count) == endBracket)
      {
        //the character is ending bracket.

        //increment the number of ending bracket encountered.
        ebcount++;
      }

      if(sbcount == ebcount)
      {
        //finished parsing through text.
        return text.substring(1, count);
      }
    }

    return null;
  }

  /**
   * Returns an enumeration describing the available options
   *
   * @return an enumeration of all the available options
   **/
  public Enumeration listOptions()
  {
    Vector newVector = new Vector(1);

    newVector.addElement(new Option
      ("\tSpecify the delimiters of the set.",
       "D", 1, "-D <value> (default \"|\")"));

    newVector.addElement(new Option
      ("\tSpecify the nominal values used to classify an attribute." +
       "Valid values are:\n" +
       "\t\"yesno\", \"truefalse\", and \"onezero\".\n" +
       "(default \"yesno\")",
       "N", 1, "-N <nominal_values>"));

    newVector.addElement(new Option
      ("\tSpecify the range of filtering. e.g. 'first-last', or '1-2,6,8-last'.",
       "R", 1, "-R <value> (default \"all\")"));

    newVector.addElement(new Option
      ("\tUse this flag to inverse the range.",
       "I", 1, "-I"));

    return newVector.elements();
  }

  /**
   * Parses a given list of options controlling the behaviour of this object.
   * Valid options are:<p>
   *
   * -D delimiters
   * Specify the delimiters of the set.
   * (default "|")<br>
   *
   * -N nominal_values
   * Specify the nominal values used to classify an attribute.
   * Valid values are "yesno", "truefalse", and "onezero".
   * (default "yesno")<br>
   *
   * -R range
   * Specify the range of filtering.
   * e.g. 'first-last', or '1-2,6,8-last'.
   * (default "all")
   *
   * -I
   * Use this flag to inverse the range.
   *
   * @param options the list of options as an array of strings
   * @exception Exception if an option is not supported
   **/
  public void setOptions(String[] options) throws Exception
  {
    //set delimiters.
    setDelimiters(Utils.getOption('D', options));

    //set nominal values.
    String nom_values = Utils.getOption('N', options);

    if(nom_values.length() != 0)
    {
      if(nom_values.equals("yesno"))
      {
        setNominalValues(new SelectedTag(YESNO, TAGS_NOMINAL_VALUES));
      }
      else if(nom_values.equals("truefalse"))
      {
        setNominalValues(new SelectedTag(TRUEFALSE, TAGS_NOMINAL_VALUES));
      }
      else if(nom_values.equals("onezero"))
      {
        setNominalValues(new SelectedTag(ONEZERO, TAGS_NOMINAL_VALUES));
      }
    }

    //set range.
    setRange(Utils.getOption('R', options));

    //set whether to inverse the range
    setRangeInverse(Utils.getFlag('I', options));

  }

  /**
   * Gets the current settings of the filter.
   *
   * @return an array of strings suitable for passing to setOptions
   **/
  public String [] getOptions()
  {
    String [] options = new String [8];
    int current = 0;

    //get delimiters.
    options[current++] = "-D";
    options[current++] = getDelimiters();

    //get nominal values.
    options[current++] = "-N";

    int nom_ID = getNominalValues().getSelectedTag().getID();

    if(nom_ID == YESNO)
    {
      options[current++] = "yesno";
    }
    else if(nom_ID == TRUEFALSE)
    {
      options[current++] = "truefalse";
    }
    else if(nom_ID == ONEZERO)
    {
      options[current++] = "onezero";
    }

    //get range.
    options[current++] = "-R";
    options[current++] = getRange();

    //get whether range should be inversed.
    if(getRangeInverse())
    {
      options[current++] = "-I";
    }

    while(current < options.length)
    {
      options[current++] = "";
    }
    
    return options;
  }

  /**
   * Main method for testing this class.
   *
   * @param argv should contain arguments to the filter: use -h for help
   **/
  public static void main(String [] argv)
  {
    try
    {
      if(Utils.getFlag('b', argv))
      {
 	      Filter.batchFilterFile(new AttributeSetFilter(), argv);
      }
      else
      {
	      Filter.filterFile(new AttributeSetFilter(), argv);
      }
    }
    catch(Exception ex)
    {
      ex.printStackTrace();
      System.out.println(ex.getMessage());
    }
  }
}




